# coding: utf-8
from urllib.parse import urljoin

from django.apps import apps
from django.urls import reverse
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.conf import settings


class Command(BaseCommand):
    help = 'Invite user to activate OTP'

    def add_arguments(self, parser):
        parser.add_argument(
            'emails', type=str, help='choosen emails, separated by a comma'
        )

    def handle(self, *args, **options):
        UserModel = get_user_model()
        concrete = apps.get_app_config('concrete')
        EmailModel = concrete.models['email']
        emails_str = options['emails'].lower()
        emails = set(emails_str.split(','))

        base_url = f'{settings.SCHEME}://{settings.HOSTNAME}'
        if settings.DEBUG is True:
            base_url += f':{settings.PORT}'
        configure_otp_url = urljoin(base_url, reverse('configure-otp'))
        email_body = settings.SEND_OTP_CONFIGURE_LINK.format(
            platform=settings.PLATFORM_NAME, link=configure_otp_url
        )
        not_found_users = []
        emails_not_sent = []
        has_error = False
        for email in emails:
            try:
                user = UserModel.objects.get(email=email)
            except UserModel.DoesNotExist:
                not_found_users.append(email)
                has_error = True
                continue

            email_instance = EmailModel(
                created_by=user,
                subject=settings.OTP_CONFIGURE_EMAIL_SUBJECT,
                body=email_body,
                receiver=user,
            )

            email_instance.save()
            if email_instance.resource_status == 'send-error':
                emails_not_sent.append(email)
                has_error = True
        if has_error is False:
            print('Emails have been sent')
        if len(not_found_users) > 0:
            emails_list_msg = ", ".join(not_found_users)
            print(f'The following emails do not exist: {emails_list_msg}')
        if len(emails_not_sent) > 0:
            emails_list_msg = ", ".join(emails_not_sent)
            print(
                'An error occured while attempting to send email '
                f'to {emails_list_msg}'
            )



# coding: utf-8
from django.conf import settings
from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from django.contrib.auth.admin import UserAdmin

from concrete_datastore.admin.admin_form import (
    MyChangeUserForm,
    MyCreationUserForm,
)
from concrete_datastore.concrete.models import divider_field_name
from django.contrib.gis.admin import OSMGeoAdmin


class MetaUserAdmin(UserAdmin, OSMGeoAdmin):
    display_wkt = True
    models_fields = ()
    readonly_fields = ['uid', 'creation_date', 'modification_date']
    date_fields = ('creation_date', 'modification_date')
    subscriptions_fields = ('unsubscribe_all', 'unsubscribe_to')
    user_fields = (
        'uid',
        'email',
        'first_name',
        'last_name',
        'password',
        'user_level',
    )
    divider_field_name_plural = divider_field_name + 's'
    form = MyChangeUserForm
    add_form = MyCreationUserForm
    add_form_template = 'admin/add_form.html'
    add_fieldsets = (
        (
            None,
            {
                'classes': ('wide',),
                'fields': ('email', 'password1', 'password2'),
            },
        ),
    )
    fieldsets = (
        (None, {'fields': user_fields}),
        (_('Scope'), {'fields': (divider_field_name_plural,)}),
        (_('Dates'), {'fields': date_fields}),
        (_('Subscriptions'), {'fields': subscriptions_fields}),
    )
    if settings.ADMIN_SHOW_USER_PERMISSIONS is True:
        fieldsets += ((_('Permissions'), {'fields': ('user_permissions',)}),)
    filter_horizontal = (divider_field_name_plural, 'unsubscribe_to')
    ordering = ('email',)

    def save_model(self, request, obj, form, change):
        if change:
            # Seul le model user a un user_level
            if 'user_level' in form.cleaned_data:
                obj.email = obj.email.lower()
                obj.set_level(
                    level=form.cleaned_data.get('user_level'), commit=False
                )

        confirmation = obj.get_or_create_confirmation()
        if settings.AUTH_CONFIRM_EMAIL_ENABLE is False:
            confirmation.confirmed = True
            confirmation.save()
        else:
            if confirmation.link_sent is False:
                confirmation.send_link()
        super(MetaUserAdmin, self).save_model(request, obj, form, change)


class MetaAdmin(OSMGeoAdmin):
    display_wkt = True
    models_fields = ()
    date_fields = ('creation_date', 'modification_date')
    date_hierarchy = 'creation_date'
    readonly_fields = [
        'uid',
        'creation_date',
        'modification_date',
        'created_by',
    ]
    permissions = (
        'public',
        'created_by',
        ('can_view_users', 'can_view_groups'),
        ('can_admin_users', 'can_admin_groups'),
    )
    fieldsets = (
        (_('Dates'), {'fields': date_fields}),
        (_('Permissions'), {'fields': permissions}),
    )
    filter_horizontal = (
        'can_view_users',
        'can_view_groups',
        'can_admin_users',
        'can_admin_groups',
    )
    # En rajoutant un css sous la forme ci-dessous, on peut rajouter du css
    # à un model admin spécifique
    # css = {
    #     'all': (
    #         'admin/css/reset_css.css',
    #     )
    # }

    class Media:
        css = {}

    def save_model(self, request, obj, form, change):
        if change is False:
            obj.created_by = request.user
        super(MetaAdmin, self).save_model(request, obj, form, change)



# coding: utf-8
import uuid
import functools
import re
from django.db.models import Q
from django.contrib.gis.measure import D
from django.contrib.auth import get_user_model
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.exceptions import ValidationError
from rest_framework.filters import BaseFilterBackend
from concrete_datastore.api.v1.datetime import format_datetime, ensure_pendulum
from django.contrib.gis.geos import Point

# target_field type is date or datetime or int or float
RANGEABLE_TYPES = (
    'DateTimeField',
    'DateField',
    'DecimalField',
    'IntegerField',
    'FloatField',
)

TYPES_VALUES_MAP = {
    "str": lambda x: x,
    "int": lambda x: int(x),
    "float": lambda x: float(x),
    "bool": lambda x: True if x.lower() == 'true' else False,
    "null": lambda x: None,
}

JSON_FILTER_PATTERN = r'^\"(?P<str>.*)\"$|(?P<int>^\d+$)|(?P<float>^\d+\.\d+([e][+-]?\d+)?$)|(?P<bool>^true$|^false$)|(?P<null>^null$|^none$)'
REGEX_DATETIME_MICROSECOND = "^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{1,6}Z$"
REGEX_DATETIME_SECOND = "^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$"


def cast_value_to_right_type(query_value):
    match = re.search(JSON_FILTER_PATTERN, query_value, flags=re.IGNORECASE)
    if match is None:
        raise ValueError(f'{query_value} is not a valid value')
    match_group = match.groupdict()
    q_type, q_value = next(
        (x, y) for x, y in match_group.items() if y is not None
    )
    return TYPES_VALUES_MAP[q_type](q_value)


def get_filter_field_type(model_class, param) -> str:
    """
    Return the type (as String) of the target field that we want to filter
    """
    splitted_param = param.split('__')
    if (
        model_class._meta.get_field(splitted_param[0]).get_internal_type()
        == 'JSONField'
    ):
        return 'JSONField'
    if len(splitted_param) == 1:
        return model_class._meta.get_field(
            splitted_param[0]
        ).get_internal_type()
    elif len(splitted_param) == 2:
        first_param, second_param = splitted_param
        if (
            model_class._meta.get_field(first_param).get_internal_type()
            != 'ForeignKey'
        ):
            raise ValidationError(
                {
                    "message": (
                        f"{param}: Multi level filters"
                        " available only for foreign keys"
                    )
                }
            )
        return (
            model_class._meta.get_field(first_param)
            .remote_field.model._meta.get_field(second_param)
            .get_internal_type()
        )

    # If we have field__fkfield__fkfieldfkfield or more, raise an error
    # We can at most have field__fkfield as filter
    else:
        raise ValidationError(
            {
                "message": (
                    f"{param}: filter not available for more than 1 level"
                )
            }
        )


def ensure_uuid_valid(value, version=None):
    try:
        uuid.UUID(value, version=version)
    except ValueError:
        return False
    else:
        return True


def convert_type(string, field_type, close_period=True):
    if string == '':
        message = "Attempting to convert an empty string to a date format"
        raise ValidationError({"message": message})
    if field_type == 'DateTimeField':
        # Expected format YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ]
        regex_ms = re.compile(REGEX_DATETIME_MICROSECOND)
        regex_second = re.compile(REGEX_DATETIME_SECOND)

        check_microseconds = regex_ms.match(str(string))
        check_seconds = regex_second.match(str(string))

        dt = ensure_pendulum(string)
        #: If the given value contains microseconds,
        #: the method convert_type must return the exact value
        if check_microseconds:
            return format_datetime(dt)

        #: If the given value does not contain microseconds but
        #: conains a datetime, the method convert_type must
        #: either return the `start_of` or `end_of` second, depending
        #: on the argument `close_period`
        #: Otherwise the return value should be the `start_of` or
        #: `end_of` day of the given date.
        #: the variable `time_limit_unit` will contain either
        #: `"second"` or `"day"`
        if check_seconds:
            time_limit_unit = 'second'
        else:
            time_limit_unit = 'day'
        if close_period is True:
            dt = dt.end_of(time_limit_unit)
        else:
            dt = dt.start_of(time_limit_unit)
        return format_datetime(dt)

    if field_type == 'DateField':
        # Expected YYYY-MM-DD
        dt = ensure_pendulum(string)
        # Deactivated by lco, a date is a date, no time in it
        # if close_period:
        #     dt.end_of('day')
        return dt.to_date_string()
    if field_type in ('DecimalField', 'FloatField'):
        return float(string)
    #: Otherwise, the field is an IntegerField
    return int(string)


class CustomShemaOperationParameters:
    def return_if_not_details(self, view, value, extra_condition=True):
        if view.detail is False and extra_condition:
            return value
        return []

    def get_q_objects(self, q_filter, q_exclude, custom_filter, exclude):
        if exclude is False:
            if q_filter is None:
                q_filter = Q(**custom_filter)
            else:
                q_filter &= Q(**custom_filter)
        else:
            if q_exclude is None:
                q_exclude = Q(**custom_filter)
            else:
                q_exclude &= Q(**custom_filter)
        return q_filter, q_exclude

    def get_custom_filtered_queryset(self, qs, q_filter, q_exclude):
        if q_filter is not None:
            qs = qs.filter(q_filter)
        if q_exclude is not None:
            qs = qs.exclude(q_exclude)
        return qs


class ExcludeFilterBackend(DjangoFilterBackend):
    """
    This class inherits form DjangoFilterBackend and uses only negated
    query_parms (only ending with '!')
    The filter method will exclude the result of super().filter_queryset
    The exclusion must be performed on each queryparam
    """

    def get_negated_query_params(self, request, view):
        filterset_fields = getattr(view, 'filterset_fields', ())
        query_params = request.query_params
        return [
            {param[:-1]: value}
            for param, value in query_params.items()
            if param.endswith('!') is True
            and '__' not in param  #: this is handeled by another class
            and '_uid' not in param  #: this is handeled by another class
            and param[:-1] in filterset_fields
        ]

    def filter_queryset(self, request, queryset, view):
        def _get_unitary_filterset_backend(query_param):
            class UnitaryFilterBacked(DjangoFilterBackend):
                def get_filterset_kwargs(self, request, queryset, view):
                    return {
                        'data': query_param,
                        'queryset': queryset,
                        'request': request,
                    }

            return UnitaryFilterBacked().filter_queryset(
                request, queryset, view
            )

        for query_param in self.get_negated_query_params(request, view):
            filtered_qs = _get_unitary_filterset_backend(
                query_param=query_param
            )
            filtered_qs_pks = filtered_qs.values_list('pk', flat=True)
            queryset = queryset.exclude(pk__in=filtered_qs_pks)
        return queryset


class FilterDistanceBackend(BaseFilterBackend, CustomShemaOperationParameters):
    suffixes_map = {
        'gte': (
            'get the values greater than or equal to the given distance '
            '(expected value format: DISTANCE,LONGITUDE,LATITUDE)'
        ),
        'lte': (
            'get the values less than or equal to the given distance '
            '(expected value format: DISTANCE,LONGITUDE,LATITUDE)'
        ),
        'gt': (
            'get the values greater than the given distance '
            '(expected value format: DISTANCE,LONGITUDE,LATITUDE)'
        ),
        'lt': (
            'get the values less than the given distance '
            '(expected value format: DISTANCE,LONGITUDE,LATITUDE)'
        ),
        'range': (
            'get the values between the two given distances (expected '
            'value format: DISTANCE1,DISTANCE2,LONGITUDE,LATITUDE)'
        ),
        'range!': (
            'get the values outside of the two given distances (expected '
            'value format: DISTANCE1,DISTANCE2,LONGITUDE,LATITUDE)'
        ),
    }

    def remove_from_queryset(self, view):
        #: Remove PointField field from filterset_fields because
        #: they cannot be filtered with the other filter backends
        filterset_fields = getattr(view, 'filterset_fields', ())
        new_filterset_fields = tuple(
            filter_field
            for filter_field in filterset_fields
            if (
                get_filter_field_type(view.model_class, filter_field)
                != 'PointField'
            )
        )
        setattr(view, 'filterset_fields', new_filterset_fields)

    def get_schema_operation_parameters(self, view):
        params = []
        for key, description in self.suffixes_map.items():
            params.extend(
                [
                    {
                        'name': f'{field_name}_distance_{key}',
                        'required': False,
                        'in': 'query',
                        'description': description,
                        'schema': {'type': 'string'},
                    }
                    for field_name in getattr(view, 'filterset_fields', ())
                    if get_filter_field_type(view.model_class, field_name)
                    == 'PointField'
                ]
            )

        self.remove_from_queryset(view=view)

        return self.return_if_not_details(view=view, value=params)

    def get_float_or_error(self, value):
        try:
            return float(value)
        except Exception:
            raise ValidationError(f'"{value}" is not a valid float')

    def filter_queryset(self, request, queryset, view):
        # Only applicable on PointField objects
        q_object_filter = None
        q_object_exclude = None
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())
        for param in query_params:
            exclude = False
            valid_param = any(
                [
                    param.endswith(f'__distance_{lookup}')
                    for lookup in self.suffixes_map.keys()
                ]
            )
            if valid_param is False:
                continue
            param_field, lookup = param.rsplit('__distance_', 1)
            if param_field not in filterset_fields:
                continue
            if (
                get_filter_field_type(queryset.model, param_field)
                != 'PointField'
            ):
                continue
            values = query_params.get(param).split(',')
            comparaison_lookup = lookup in ('lt', 'lte', 'gt', 'gte')
            range_lookup = lookup in ('range', 'range!')
            if comparaison_lookup is True:
                #: If the lookup is one of [lt, lte, gt, gte]
                #: the split should have three elements
                if len(values) != 3:
                    raise ValidationError(
                        f"Distance filter with lookup {lookup} needs the "
                        "following parameters: Distance, longitude and "
                        "latitude"
                    )
                distance = self.get_float_or_error(values[0])
                longitude = self.get_float_or_error(values[1])
                latitude = self.get_float_or_error(values[2])
                point = Point(longitude, latitude)
                custom_filter = {param: (point, D(m=distance))}
            elif range_lookup is True:
                #: If the lookup is one of [range, range!]
                #: the split should have four elements
                if len(values) != 4:
                    raise ValidationError(
                        f"Distance filter with lookup {lookup} needs the "
                        "following parameters: Distance1, Distance2, "
                        "longitude and latitude"
                    )
                distance1 = self.get_float_or_error(values[0])
                distance2 = self.get_float_or_error(values[1])
                min_distance = min(distance1, distance2)
                max_distance = max(distance1, distance2)
                longitude = self.get_float_or_error(values[2])
                latitude = self.get_float_or_error(values[3])
                point = Point(longitude, latitude)
                custom_filter = {
                    '{}__distance_gte'.format(param_field): (
                        point,
                        D(m=min_distance),
                    ),
                    '{}__distance_lte'.format(param_field): (
                        point,
                        D(m=max_distance),
                    ),
                }
                if lookup.endswith('!'):
                    exclude = True
            else:
                raise ValidationError(
                    f"Distance filter with lookup {lookup} is not supported. "
                    f"Supported lookup values are {list(self.suffixes_map)}"
                )

            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )

        self.remove_from_queryset(view=view)
        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        )


class FilterUserByLevel(BaseFilterBackend, CustomShemaOperationParameters):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': 'level',
                'required': False,
                'in': 'query',
                'schema': {'type': 'string'},
            },
            {
                'name': 'level!',
                'required': False,
                'in': 'query',
                'schema': {'type': 'string'},
            },
            {
                'name': 'atleast',
                'required': False,
                'in': 'query',
                'schema': {'type': 'string'},
            },
            {
                'name': 'atleast!',
                'required': False,
                'in': 'query',
                'schema': {'type': 'string'},
            },
        ]
        return self.return_if_not_details(
            view=view,
            value=params,
            extra_condition=view.model_class == get_user_model(),
        )

    def filter_queryset(self, request, queryset, view):
        # Only appyable on user models
        if view.basename.lower() != 'user':
            return queryset

        query_params = request.query_params
        if 'level' in query_params:
            return queryset.model.filter_by_exact_level(
                queryset=queryset, level=query_params.get('level')
            )
        if 'level!' in query_params:
            return queryset.model.exclude_by_exact_level(
                queryset=queryset, level=query_params.get('level!')
            )
        if 'atleast' in query_params:
            return queryset.model.filter_by_at_least_level(
                queryset=queryset, level=query_params.get('atleast')
            )
        if 'atleast!' in query_params:
            return queryset.model.exclude_by_at_least_level(
                queryset=queryset, level=query_params.get('atleast!')
            )

        return queryset


class FilterSupportingOrBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__in{neg}',
                'required': False,
                'in': 'query',
                'description': (
                    'List of values separated by comma'
                    f'{" (to exclude)" if neg == "!" else ""}'
                ),
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if get_filter_field_type(view.model_class, field_name)
            != 'ManyToManyField'
            for neg in ('', '!')
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object_filter = None
        q_object_exclude = None
        for param in query_params:
            exclude = False
            bare_param = param
            if param.endswith('!'):
                bare_param = param[:-1]
                exclude = True

            if not bare_param.endswith('__in'):
                continue
            if bare_param.replace('__in', '') not in filterset_fields:
                continue
            values = query_params.get(param).split(',')

            filter_field_type = get_filter_field_type(
                queryset.model, bare_param.replace('__in', '')
            )
            if filter_field_type in (
                'UUIDField',
                'ForeignKey',
                'ManyToManyField',
            ):
                for value in values:
                    if not ensure_uuid_valid(value):
                        message = f"'{value}' is not a valid UUID"
                        raise ValidationError(
                            {"message": (f"{bare_param}: {message}")}
                        )

            if filter_field_type == 'BooleanField':
                if set(values).difference(['True', 'False', 'None']):
                    message = (
                        f"{bare_param}: {values} must contain olny 'True', "
                        "'False' and/or 'None' (case sensitive)"
                    )
                    raise ValidationError({"message": message})
            custom_filter = {bare_param: values}

            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )

        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        )


class FilterJSONFieldsBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def remove_from_queryset(self, view):
        #: Remove JSONField field from filterset_fields because
        #: they cannot be filtered with the other filter backends
        filterset_fields = getattr(view, 'filterset_fields', ())
        new_filterset_fields = tuple(
            filter_field
            for filter_field in filterset_fields
            if (
                get_filter_field_type(view.model_class, filter_field)
                != 'JSONField'
            )
        )
        setattr(view, 'filterset_fields', new_filterset_fields)

    def get_schema_operation_parameters(self, view):
        #: The json filter is generic and has no rules. So it is removed
        #: from the openAPI schema.
        #: In this method we will juste remove the JSON fields from the
        #: filterset_fields of the view and return an empty list

        self.remove_from_queryset(view=view)
        return []

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object = None
        self.remove_from_queryset(view=view)
        q_object_filter = None
        q_object_exclude = None
        for param in query_params:
            exclude = False
            bare_param = param
            if param.endswith('!'):
                bare_param = param[:-1]
                exclude = True

            splitted_query_params = bare_param.split('__')
            if len(splitted_query_params) == 1:
                continue
            param_field_name = splitted_query_params[0]

            if param_field_name not in filterset_fields:
                continue
            if (
                get_filter_field_type(queryset.model, param_field_name)
                != 'JSONField'
            ):
                continue

            value = query_params.get(param)
            try:
                custom_filter = {bare_param: cast_value_to_right_type(value)}
            except ValueError as e:
                raise ValidationError({'message': f'"{param}": {e}'})
            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )
            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)

        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        )


class FilterSupportingContainsBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__contains{neg}',
                'required': False,
                'description': f'String{" (to exclude)" if neg == "!" else ""}',
                'in': 'query',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if get_filter_field_type(view.model_class, field_name)
            in ('CharField', 'TextField')
            for neg in ('', '!')
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object_filter = None
        q_object_exclude = None
        for param in query_params:
            exclude = False
            bare_param = param
            if param.endswith('!'):
                bare_param = param[:-1]
                exclude = True

            if not bare_param.endswith('__contains'):
                continue
            param_field = bare_param.replace('__contains', '')
            if param_field not in filterset_fields:
                continue
            if not get_filter_field_type(queryset.model, param_field) in (
                'CharField',
                'TextField',
            ):
                continue

            custom_filter = {bare_param: query_params.get(param)}
            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )

        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        )


class FilterSupportingInsensitiveContainsBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__icontains{neg}',
                'required': False,
                'description': f'String{" (to exclude)" if neg == "!" else ""}',
                'in': 'query',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if get_filter_field_type(view.model_class, field_name)
            in ('CharField', 'TextField')
            for neg in ('', '!')
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object_filter = None
        q_object_exclude = None
        for param in query_params:
            exclude = False
            bare_param = param
            if param.endswith('!'):
                bare_param = param[:-1]
                exclude = True

            if not bare_param.endswith('__icontains'):
                continue
            param_field = bare_param.replace('__icontains', '')
            if param_field not in filterset_fields:
                continue
            if not get_filter_field_type(queryset.model, param_field) in (
                'CharField',
                'TextField',
            ):
                continue

            custom_filter = {bare_param: query_params.get(param)}

            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )

        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        )


class FilterSupportingEmptyBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__isempty',
                'required': False,
                'in': 'query',
                'description': 'True or False',
                'schema': {'type': 'boolean'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if get_filter_field_type(view.model_class, field_name)
            in ('CharField', 'TextField')
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object_filter = None
        q_object_exclude = None

        for param in query_params:
            exclude = False

            if not param.endswith('__isempty'):
                continue
            if query_params.get(param).lower() == 'false':
                exclude = True
            elif query_params.get(param).lower() != 'true':
                continue
            param = param.replace('__isempty', '')
            if param not in filterset_fields:
                continue
            if get_filter_field_type(queryset.model, param) in (
                'CharField',
                'TextField',
            ):
                custom_filter = {'{}__exact'.format(param): ''}
            else:
                continue
            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )
        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        )


class FilterSupportingRangeBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__range{neg}',
                'required': False,
                'in': 'query',
                'description': (
                    'A range of values separated by comma'
                    f'{" (to exclude)" if neg == "!" else ""}'
                ),
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            + ('creation_date', 'modification_date')
            if get_filter_field_type(view.model_class, field_name)
            in RANGEABLE_TYPES
            for neg in ('', '!')
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ()) + (
            'creation_date',
            'modification_date',
        )

        q_object_filter = None
        q_object_exclude = None

        for param in query_params:
            exclude = False
            bare_param = param
            if param.endswith('!'):
                bare_param = param[:-1]
                exclude = True
            if not param.endswith('__range'):
                continue
            target_field = bare_param.replace('__range', '')
            if target_field not in filterset_fields:
                continue

            target_field_type = get_filter_field_type(
                queryset.model, target_field
            )
            if target_field_type not in RANGEABLE_TYPES:
                continue

            values = query_params.get(param).split(',')

            if len(values) < 2:
                raise ValidationError(
                    {
                        "message": (
                            "A comma is expected in the value of the filter. "
                            "Expected values are '<date1>,<date2>', '<date1>,'"
                            " or ',<date2>'"
                        )
                    }
                )
            if len(values) > 2:
                raise ValidationError(
                    {
                        "message": (
                            'Only two comma-separated values are expected, '
                            f'got {len(values)}: {values}'
                        )
                    }
                )

            range_start, range_end = values

            if range_start == '' and range_end == '':
                continue

            elif range_start != '' and range_end == '':
                param = bare_param.replace('__range', '__gte')
                values = convert_type(
                    range_start, target_field_type, close_period=False
                )

            elif range_start == '' and range_end != '':
                param = bare_param.replace('__range', '__lte')
                values = convert_type(
                    range_end, target_field_type, close_period=True
                )

            else:
                values = (
                    convert_type(
                        range_start, target_field_type, close_period=False
                    ),
                    convert_type(
                        range_end, target_field_type, close_period=True
                    ),
                )

            custom_filter = {param: values}
            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )

        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        )


class FilterSupportingComparaisonBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        suffixes_map = {
            'gte': 'get the values greater than or equal to a given value',
            'lte': 'get the values less than or equal to a given value',
            'gt': 'get the values greater than to a given value',
            'lt': 'get the values less than to a given value',
        }
        params = []
        for key, description in suffixes_map.items():
            params.extend(
                [
                    {
                        'name': f'{field_name}__{key}',
                        'required': False,
                        'in': 'query',
                        'description': f'{description}',
                        'schema': {'type': 'string'},
                    }
                    for field_name in getattr(view, 'filterset_fields', ())
                    + ('creation_date', 'modification_date')
                    if get_filter_field_type(view.model_class, field_name)
                    in RANGEABLE_TYPES
                ]
            )
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ()) + (
            'creation_date',
            'modification_date',
        )

        def get_param_from_query(param):
            if param.endswith('__gte'):
                return param.replace('__gte', '')
            if param.endswith('__lte'):
                return param.replace('__lte', '')
            if param.endswith('__gt'):
                return param.replace('__gt', '')
            if param.endswith('__lt'):
                return param.replace('__lt', '')
            return None

        q_object_filter = None

        def get_custom_filters(param):
            target_field = get_param_from_query(param)
            if target_field not in filterset_fields:
                return None

            target_field_type = get_filter_field_type(
                queryset.model, target_field
            )
            if target_field_type not in RANGEABLE_TYPES:
                return None
            close_period = True
            if param.endswith('__lt') or param.endswith('__gte'):
                close_period = False
            value = convert_type(
                query_params.get(param),
                target_field_type,
                close_period=close_period,
            )
            if value is None:
                return None
            return {param: value}

        include_filters = [
            get_custom_filters(qp)
            for qp in query_params
            if get_custom_filters(qp) is not None
        ]
        q_object_filter = functools.reduce(
            lambda a, b: a & Q(**b), include_filters, Q()
        )
        return queryset.filter(q_object_filter)


class FilterSupportingForeignKey(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}_uid{neg}',
                'required': False,
                'in': 'query',
                'description': (
                    f'UID of the FK{" (to exclude)" if neg == "!" else ""}'
                ),
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if get_filter_field_type(view.model_class, field_name)
            == 'ForeignKey'
            for neg in ('', '!')
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())
        q_object_filter = None
        q_object_exclude = None

        for param in query_params:
            exclude = False
            bare_param = param
            if param.endswith('!'):
                bare_param = param[:-1]
                exclude = True
            if not bare_param.endswith('_uid'):
                continue
            cleaned_param = bare_param.replace('_uid', '')
            if cleaned_param not in filterset_fields:
                continue
            cleaned_param_type = get_filter_field_type(
                queryset.model, cleaned_param
            )
            if not cleaned_param_type == 'ForeignKey':
                continue
            value = query_params.get(param)
            custom_filter = {cleaned_param: value}
            #:  "value" must be a valid UUID4, otherwise raise ValidationError
            #:  raises ValueError if not UUID4
            if not ensure_uuid_valid(value, version=4):
                message = f'{bare_param}: « {value} » is not a valid UUID'
                raise ValidationError({"message": message})
            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )

        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        )


class FilterForeignKeyIsNullBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__isnull',
                'required': False,
                'in': 'query',
                'schema': {'type': 'boolean'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if get_filter_field_type(view.model_class, field_name)
            in ('ForeignKey', 'ManyToManyField')
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object = None
        for param in query_params:
            if not param.endswith('__isnull'):
                continue
            if query_params.get(param) not in ['true', 'false']:
                continue
            field_name = param.replace('__isnull', '')
            if field_name not in filterset_fields:
                continue
            param_value = True if query_params.get(param) == 'true' else False
            cleaned_param_type = get_filter_field_type(
                queryset.model, field_name
            )
            if cleaned_param_type in ('ForeignKey', 'ManyToManyField'):
                custom_filter = {param: param_value}
            else:
                continue

            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)
        if q_object is None:
            return queryset

        return queryset.filter(q_object)


class FilterSupportingManyToMany(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__in{neg}',
                'required': False,
                'description': 'List of uids separated by comma',
                'in': 'query',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if get_filter_field_type(view.model_class, field_name)
            == 'ManyToManyField'
            for neg in ('', '!')
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())
        q_object_filter = None
        q_object_exclude = None

        for param in query_params:
            exclude = False
            bare_param = param
            if param.endswith('!'):
                bare_param = param[:-1]
                exclude = True
            if not bare_param.endswith('__in'):
                continue
            field_name = bare_param.replace('__in', '')
            if field_name not in filterset_fields:
                continue
            cleaned_param_type = get_filter_field_type(
                queryset.model, field_name
            )
            if not cleaned_param_type == 'ManyToManyField':
                continue
            values = set(query_params.get(param).split(','))

            #:  "value" must be a valid UUID4, otherwise raise ValidationError
            for value in values:
                if not ensure_uuid_valid(value, version=4):
                    #:  raises ValueError if not UUID4
                    message = f'{bare_param}: « {value} » is not a valid UUID'
                    raise ValidationError({"message": message})

            custom_filter = {bare_param: values}
            q_object_filter, q_object_exclude = self.get_q_objects(
                q_filter=q_object_filter,
                q_exclude=q_object_exclude,
                custom_filter=custom_filter,
                exclude=exclude,
            )

        return self.get_custom_filtered_queryset(
            qs=queryset, q_filter=q_object_filter, q_exclude=q_object_exclude
        ).distinct()



# coding: utf-8

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from concrete_datastore.concrete.management.commands.reset_password import (
    generate_random_password,
    get_admin_url,
)


class Command(BaseCommand):
    help = 'Create users'

    def add_arguments(self, parser):
        parser.add_argument('email', type=str, help='Super user email')

    def handle(self, *args, **options):
        concrete = apps.get_app_config('concrete')
        EmailModel = concrete.models['email']
        UserModel = get_user_model()

        email = options['email'].lower()

        user, created = UserModel.objects.get_or_create(email=email)

        if created:
            password = generate_random_password()
            user.set_password(password)
            user.set_level('superuser')
            user.save()

            admin_url = get_admin_url()
            email_instance = EmailModel(
                created_by=user,
                subject='Access to Concrete Instance',
                body=settings.CREATE_SUPERUSER_EMAIL_MESSAGE_BODY.format(
                    hostname=settings.HOSTNAME,
                    admin_url=admin_url,
                    email=email,
                    password=password,
                ),
                receiver=user,
            )

            email_instance.save()



# coding: utf-8
import uuid
import warnings
import logging
from copy import deepcopy
from urllib.parse import urljoin

from django.db import models
from django.core import validators
from django.http import HttpRequest
from django.conf import settings
from django.utils.encoding import force_str
from django.core.exceptions import PermissionDenied

from rest_framework import exceptions, serializers
from rest_framework.fields import empty
from rest_framework.compat import uritemplate
from rest_framework.request import Request
from rest_framework.schemas.utils import get_pk_description
from rest_framework.schemas.openapi import AutoSchema as AutoSchemaSuper
from rest_framework.schemas.generators import (
    EndpointEnumerator,
    BaseSchemaGenerator as BaseSchemaSuper,
)

import concrete_datastore

logger = logging.getLogger('open-api-spec')

CONCRETE_FAIL_RESPONSES = {
    '404': {'description': 'Not found'},
    '400': {'description': 'Bad request'},
    '401': {'description': 'Unauthorized'},
    '403': {'description': 'Forbidden'},
    '429': {'description': 'Too many requests'},
    '412': {'description': 'Precondition failed'},
}

CONCRETE_DEFAULT_LIST_RESP = {
    'objects_count': {'type': 'integer'},
    'next': {'type': 'string'},
    'previous': {'type': 'string'},
    'objects_count_per_page': {'type': 'integer'},
    'num_total_pages': {'type': 'integer'},
    'num_current_page': {'type': 'integer'},
    'max_allowed_objects_per_page': {'type': 'integer'},
    'model_name': {'type': 'string'},
    'model_verbose_name': {'type': 'string'},
    'list_display': {'type': 'array', 'items': {'type': 'string'}},
    'list_filter': {'type': 'array', 'items': {'type': 'object'}},
    'total_objects_count': {'type': 'integer'},
    'create_url': {'type': 'string'},
}

SWAGGER_SETTINGS = {
    'APP_VERSION': concrete_datastore.__version__,
    'DESCRIPTION': 'Concrete datastore API.',
    'CONTACT': {'email': 'contact@netsach.org'},
    'TERMS_OF_SERVICE': '',
    'SECURITY_DEFINITIONS': {
        'Token': {'type': 'apiKey', 'in': 'header', 'name': 'Authorization'}
    },
}

DEFAULT_LIST_VIEW_PARAMETERS = [
    {
        'name': f'c_resp_nested',
        'required': False,
        'description': 'Whether the response should be nested or not',
        'in': 'query',
        'schema': {'type': 'boolean'},
    },
    {
        'name': f'c_resp_page_size',
        'required': False,
        'description': (
            'Max objects to dosplay in a page: '
            f'number between 1 and {settings.DEFAULT_PAGE_SIZE}'
        ),
        'in': 'query',
        'schema': {'type': 'integer'},
    },
    {
        'name': f'timestamp_start',
        'required': False,
        'description': 'Starting timestamp of the request',
        'in': 'query',
        'schema': {'type': 'float'},
    },
    {
        'name': f'timestamp_end',
        'required': False,
        'description': 'Ending timestamp of the request',
        'in': 'query',
        'schema': {'type': 'float'},
    },
]

FIELD_TYPES_MAP = {
    "IntegerField": "integer",
    "BooleanField": "boolean",
    "FloatField": "float",
    "DeciamlField": "float",
}


class InvalidTokenUser(Exception):
    pass


class HasPermissionFakeUser:
    def __init__(self, level):
        level = level.lower().strip()
        self.pk = uuid.uuid4()
        self.is_anonymous = level == 'blocked'
        self.is_authenticated = self.is_anonymous is False
        self.is_superuser = level == 'superuser'
        self.admin = self.is_superuser or level == 'admin'
        self.is_staff = self.admin or level == 'manager'
        self.is_active = self.is_staff or level == 'simpleuser'
        self.is_at_least_admin = self.admin
        self.is_at_least_staff = self.admin or self.is_staff

    def get_roles(self):
        if self.is_anonymous is False:
            logger.info(
                f'Project requires to access User Roles, but we are'
                ' trying to generate the specification without any'
                ' Database for a user with "{self.level}" level.'
            )
        return []

    @property
    def level(self):
        if self.is_superuser:
            return 'superuser'
        if self.admin:
            return 'admin'
        if self.is_staff:
            return 'manager'
        if self.is_active:
            return 'simpleuser'
        return 'blocked'

    def is_confirmed(self):
        return self.is_active

    def get_level(self):
        if self.is_superuser:
            return 'SuperUser'
        if self.admin:
            return 'Admin'
        if self.is_staff:
            return 'Manager'
        if self.is_active:
            return 'SimpleUser'
        return 'Blocked'


class CustomEndpointEnumerator(EndpointEnumerator):
    def get_allowed_methods(self, callback):
        """
        Return a list of the valid HTTP methods for this endpoint.
        """
        if hasattr(callback, 'actions'):
            actions = set(callback.actions)
            methods = [method.upper() for method in actions]
        else:
            #:  Some views require a api_namespace attribute
            kwargs = getattr(callback, 'view_initkwargs', {})
            methods = callback.cls(**kwargs).allowed_methods

        return [
            method for method in methods if method not in ('OPTIONS', 'HEAD')
        ]


class BaseSchemaGenerator(BaseSchemaSuper):
    endpoint_inspector_cls = CustomEndpointEnumerator
    urlconf = None
    url = None
    endpoints = None

    def __init__(
        self,
        patterns=None,
        servers=None,
        user_token=None,
        from_db=False,
        user_level='blocked',
    ):
        super().__init__(
            title=settings.OPENAPI_SPEC_TITLE,
            patterns=patterns,
            description=SWAGGER_SETTINGS['DESCRIPTION'],
        )

        license_content = getattr(settings, 'LICENSE')

        self.title = settings.OPENAPI_SPEC_TITLE
        self.license = {'name': license_content} if license_content else {}
        self.terms_of_service = SWAGGER_SETTINGS['TERMS_OF_SERVICE']
        self.version = SWAGGER_SETTINGS['APP_VERSION']
        self.security_definitions = SWAGGER_SETTINGS['SECURITY_DEFINITIONS']
        self.contact = SWAGGER_SETTINGS['CONTACT']
        self.servers = servers
        self.from_db = from_db
        self.user = self._get_user(token=user_token, level=user_level)

    def _get_user(self, token, level):
        if self.from_db is False:
            return HasPermissionFakeUser(level=level)

        #:  AuthToken is imported only if `self.from_db` is True in order
        #:  to avoid any database errors
        from concrete_datastore.concrete.models import AuthToken

        # pylint: disable=no-member
        user_token = AuthToken.objects.filter(pk=token).first()
        if user_token:
            return user_token.user
        raise PermissionDenied()

    def _get_paths_and_endpoints(self, request):
        """
        Generate (path, method, view) given (path, method, callback) for paths.
        """
        paths = []
        view_endpoints = []
        for path, method, callback in self.endpoints:
            view = self.create_view(callback, method, request)
            path = self.coerce_path(path, method, view)
            path = path.replace('\\', '')
            paths.append(path)
            view_endpoints.append((path, method, view))

        return paths, view_endpoints

    def coerce_path(self, path, method, view):
        if not self.coerce_path_pk:
            return path
        return path.replace('{pk}', '{uid}')

    def has_view_permissions(self, path, method, view):
        """
        Return `True` if the incoming request has the correct view permissions.
        """
        request = view.request
        if request is None:
            request = Request(HttpRequest())
            setattr(request, 'method', method)
            setattr(request, 'user', self.user)
            setattr(view, 'request', request)
        elif isinstance(self.user, HasPermissionFakeUser):
            setattr(request, 'user', self.user)

        for permission_cls in view.permission_classes:
            if not permission_cls().has_permission(request, view):
                return False
        return True


def filter_empty(data):
    if isinstance(data, dict):
        return {
            key: filter_empty(value) for key, value in data.items() if value
        }
    if isinstance(data, (list, set, tuple)):
        return list(map(filter_empty, data))
    else:
        return data


# Generator
class SchemaGenerator(BaseSchemaGenerator):
    def get_info(self):
        info = {
            "title": self.title,
            "description": self.description,
            "termsOfService": self.terms_of_service,
            "contact": self.contact,
            "license": self.license,
            "version": self.version,
        }
        return info

    def get_paths_and_components(self, request=None):
        result_path = {}
        result_components = {}
        operation_ids = set()

        paths, view_endpoints = self._get_paths_and_endpoints(request)

        # Only generate the path prefix for paths that will be included
        if not paths:
            return None

        for path, method, view in view_endpoints:
            if (
                not self.has_view_permissions(path, method, view)
                or '/sets{' in path
                or path.endswith('/export/')
            ):
                continue
            operation, operation_ids = view.schema.get_operation(
                path=path,
                method=method,
                from_db=self.from_db,
                operation_ids=operation_ids,
            )
            components = view.schema.get_component_schemas()
            # Normalise path for any provided mount url.
            if path.startswith('/'):
                path = path[1:]
            path = urljoin(self.url or '/', path)

            result_path.setdefault(path, {})
            result_path[path][method.lower()] = operation
            result_components.update(components)

        return result_path, result_components

    def get_schema(self, request=None, public=False):
        """
        Generate a OpenAPI schema.
        """
        self._initialise_endpoints()

        paths, components = self.get_paths_and_components(
            None if public else request
        )
        if not paths:
            return None

        schema_servers = {}
        if self.servers:
            schema_servers = {
                'servers': [{'url': server} for server in self.servers]
            }
        schema = {
            'openapi': '3.0.2',
            'info': filter_empty(self.get_info()),
            **schema_servers,
            'security': [
                {security: []} for security in self.security_definitions.keys()
            ],
            'paths': filter_empty(paths),
            'components': {
                'schemas': filter_empty(components),
                'securitySchemes': filter_empty(self.security_definitions),
            },
        }
        return schema


class AutoSchema(AutoSchemaSuper):
    components = {}

    def get_component_schemas(self):
        return self.components

    def get_operation(self, path, method, from_db, operation_ids):
        operation = {}

        op_id, operation_ids = self.get_distinct_operation_id(
            self.get_operation_id(path, method), operation_ids
        )
        operation['operationId'] = op_id
        if from_db is False:
            self.view.get_queryset = lambda: None

        parameters = []
        parameters += self.get_custom_path_parameters(path, method)
        parameters += self.get_pagination_parameters(path, method)
        parameters += self.get_custom_filter_parameters(path, method)
        operation['parameters'] = parameters

        request_body = self.get_custom_request_body(path, method)
        if request_body:
            operation['requestBody'] = request_body
        operation['responses'] = self.get_custom_responses(path, method)

        return operation, operation_ids

    def get_custom_filter_parameters(self, path, method):
        if not self.allows_filters(path, method) and "/stats" not in path:
            return []
        if getattr(self.view, 'detail', False) is True:
            return []
        parameters = []
        if "/stats" not in path:
            parameters = deepcopy(DEFAULT_LIST_VIEW_PARAMETERS)
        if hasattr(self.view, 'model_class'):
            for field in getattr(self.view, 'filterset_fields', []):
                filters_module = concrete_datastore.api.v1.filters
                field_type = filters_module.get_filter_field_type(
                    self.view.model_class, field
                )
                if field_type in ('ForeignKey', 'ManyToManyField'):
                    continue
                swagger_filed_type = FIELD_TYPES_MAP.get(field_type, 'string')
                parameters.extend(
                    [
                        {
                            'name': field,
                            'required': False,
                            'description': 'Exact value of the field',
                            'in': 'query',
                            'schema': {'type': swagger_filed_type},
                        },
                        {
                            'name': f'{field}!',
                            'required': False,
                            'description': (
                                'Exact value of the field (to exclude)'
                            ),
                            'in': 'query',
                            'schema': {'type': swagger_filed_type},
                        },
                    ]
                )
        for filter_backend in self.view.filter_backends:
            parameters += filter_backend().get_schema_operation_parameters(
                self.view
            )
        return parameters

    def get_distinct_operation_id(self, op_id, operation_ids):
        # Ensure operation Id is unique by incrementing the suffix
        if op_id not in operation_ids:
            operation_ids.add(op_id)
            return op_id, operation_ids

        inc = 1
        while f'{op_id}_{inc}' in operation_ids:
            inc += 1

        op_id = f'{op_id}_{inc}'
        operation_ids.add(op_id)
        return op_id, operation_ids

    def get_custom_path_parameters(self, path, method):
        """
        Return a list of parameters from templated path variables.
        """
        model = getattr(self.view.__class__, 'model_class', None)

        parameters = []
        for variable in uritemplate.variables(path):
            if '/stats' in path:
                parameter = {
                    "name": "var",
                    "in": "path",
                    "required": False,
                    "description": (
                        "Timestamp URL. Example:\n- \"/timestamp_start:"
                        "1603801749.315451\"\n- \"/timestamp_start:1603801749."
                        "315451-timestamp_end:1603801850.463\""
                    ),
                    'schema': {'type': 'string'},
                }
                parameters.append(parameter)
                continue
            description = ''
            if model is not None:
                # Attempt to infer a field description if possible.
                try:
                    model_field = model._meta.get_field(variable)
                except Exception:
                    model_field = None

                if model_field is not None and model_field.help_text:
                    description = force_str(model_field.help_text)
                elif model_field is not None and model_field.primary_key:
                    description = get_pk_description(model, model_field)

            parameter = {
                "name": variable,
                "in": "path",
                "required": True,
                "description": description,
                'schema': {'type': 'string'},
            }
            parameters.append(parameter)

        return parameters

    def custom_map_field(self, field):
        # Related fields.
        if isinstance(field, serializers.ManyRelatedField):
            return {
                'type': 'array',
                'items': {'type': 'string', 'format': 'uid'},
            }

        if isinstance(field, serializers.PrimaryKeyRelatedField):
            model = getattr(field.queryset, 'model', None)
            if model is not None:
                model_field = model._meta.pk
                if isinstance(model_field, models.AutoField):
                    return {'type': 'string', 'format': 'uid'}
        if isinstance(field, serializers.ListSerializer):
            comp_name = self.get_component_name(path=None, serializer=field)
            if comp_name not in self.components.keys():
                self.components.update(
                    {comp_name: self.custom_map_serializer(field.child)}
                )
            return {
                'type': 'array',
                'items': {'$ref': f'#/components/schemas/{comp_name}'},
            }
        if isinstance(field, serializers.Serializer):
            comp_name = self.get_component_name(path=None, serializer=field)
            if comp_name not in self.components.keys():
                self.components.update(
                    {comp_name: self.custom_map_serializer(field)}
                )
            return {'$ref': f'#/components/schemas/{comp_name}'}

        # Related fields.
        if isinstance(field, serializers.ManyRelatedField):
            return {
                'type': 'array',
                'items': {'type': 'string', 'format': 'uid'},
            }

        if isinstance(field, serializers.PrimaryKeyRelatedField):
            model = getattr(field.queryset, 'model', None)
            if model is not None:
                model_field = model._meta.pk
                if isinstance(model_field, models.AutoField):
                    return {'type': 'string', 'format': 'uid'}
        return super().map_field(field)

    def custom_map_serializer(self, serializer, nested=True):
        required = []
        properties = {}

        for field in serializer.fields.values():
            if (
                isinstance(field, serializers.PrimaryKeyRelatedField)
                and nested is False
                and field.field_name.endswith('_uid') is False
            ):
                continue

            if isinstance(field, serializers.HiddenField):
                continue

            if field.required:
                required.append(field.field_name)

            schema = self.custom_map_field(field)
            if field.read_only:
                schema['readOnly'] = True
            if field.write_only:
                schema['writeOnly'] = True
            if field.allow_null:
                schema['nullable'] = True
            if field.default and field.default != empty:
                schema['default'] = field.default
            if field.help_text:
                schema['description'] = str(field.help_text)
            self.custom_map_field_validators(field, schema)

            properties[field.field_name] = schema

        result = {'properties': properties}
        if len(required) > 0:
            result['required'] = required
        else:
            result['required'] = []

        return result

    def custom_map_field_validators(self, field, schema):
        #:  RegexValidator causes problems generating regex pattern
        #:  remove it for now
        for v in field.validators:
            if isinstance(v, validators.EmailValidator):
                schema['format'] = 'email'
            if isinstance(v, validators.URLValidator):
                schema['format'] = 'uri'
            elif isinstance(v, validators.MaxLengthValidator):
                attr_name = 'maxLength'
                if isinstance(field, serializers.ListField):
                    attr_name = 'maxItems'
                schema[attr_name] = v.limit_value
            elif isinstance(v, validators.MinLengthValidator):
                attr_name = 'minLength'
                if isinstance(field, serializers.ListField):
                    attr_name = 'minItems'
                schema[attr_name] = v.limit_value
            elif isinstance(v, validators.MaxValueValidator):
                schema['maximum'] = v.limit_value
            elif isinstance(v, validators.MinValueValidator):
                schema['minimum'] = v.limit_value
            elif isinstance(v, validators.DecimalValidator):
                if v.decimal_places:
                    schema['multipleOf'] = float(
                        '.' + (v.decimal_places - 1) * '0' + '1'
                    )
                if v.max_digits:
                    digits = v.max_digits
                    if v.decimal_places is not None and v.decimal_places > 0:
                        digits -= v.decimal_places
                    schema['maximum'] = int(digits * '9') + 1
                    schema['minimum'] = -schema['maximum']

    def get_custom_request_body(self, path, method):
        view = self.view

        if method not in ('PUT', 'PATCH', 'POST'):
            return {}

        if not hasattr(view, 'get_serializer'):
            return {}

        try:
            serializer = view.get_serializer()
        except exceptions.APIException:
            serializer = None
            warnings.warn(
                '{}.get_serializer() raised an exception during '
                'schema generation. Serializer fields will not be '
                'generated for {} {}.'.format(
                    view.__class__.__name__, method, path
                )
            )

        if not isinstance(serializer, serializers.Serializer):
            return {}

        nested = True
        if method in ('POST', 'PATCH', 'PUT'):
            nested = False
        content = self.custom_map_serializer(serializer, nested)
        # No required fields for PATCH
        if method == 'PATCH':
            del content['required']
        # No read_only fields for request.
        for name, schema in content['properties'].copy().items():
            if 'readOnly' in schema:
                del content['properties'][name]

        component_name = self.get_component_name(path, serializer, method)
        if component_name not in self.components.keys():
            self.components.update({component_name: content})

        return {
            'content': {
                ct: {
                    'schema': {
                        '$ref': f'#/components/schemas/{component_name}'
                    }
                }
                for ct in self.request_media_types
            }
        }

    def get_custom_responses(self, path, method):
        resp_dict = CONCRETE_FAIL_RESPONSES.copy()
        if method == 'POST':
            success_status = '201'
            success_description = 'Created'
        elif method == 'DELETE':
            return {'204': {'description': 'No content'}, **resp_dict}
        else:
            success_status = '200'
            success_description = 'Ok'
        content = {}

        if not hasattr(self.view, 'get_serializer'):
            return {'200': {'description': 'Ok', 'content': {}}}
        try:
            serializer = self.view.get_serializer()
        except exceptions.APIException:
            serializer = None
            warnings.warn(
                '{}.get_serializer() raised an exception during '
                'schema generation. Serializer fields will not be '
                'generated for {} {}.'.format(
                    self.view.__class__.__name__, method, path
                )
            )

        if isinstance(serializer, serializers.Serializer):
            content = self.custom_map_serializer(serializer)

            # No write_only fields for response.
            for name, schema in content['properties'].copy().items():
                if 'writeOnly' in schema:
                    del content['properties'][name]
                    content['required'] = [
                        f for f in content['required'] if f != name
                    ]
        item_name = self.get_component_name(path, serializer)
        if item_name not in self.components.keys():
            self.components.update({item_name: content})
        if method != 'GET' or '{uid}' in path:
            component_name = item_name
        else:
            list_name = self.get_component_name(path, serializer, method)
            if list_name not in self.components.keys():
                list_dict = CONCRETE_DEFAULT_LIST_RESP.copy()
                list_dict.update(
                    {
                        'results': {
                            'type': 'array',
                            'items': {
                                '$ref': f'#/components/schemas/{item_name}'
                            },
                        }
                    }
                )
                self.components.update(
                    {list_name: {'properties': list_dict, 'type': 'object'}}
                )
            component_name = list_name

        resp_dict.update(
            {
                success_status: {
                    'description': success_description,
                    'content': {
                        ct: {
                            'schema': {
                                '$ref': (
                                    f'#/components/schemas/{component_name}'
                                )
                            }
                        }
                        for ct in self.request_media_types
                    },
                }
            }
        )
        return resp_dict

    def get_component_name(self, path, serializer, method=None):
        if method in ('POST', 'PATCH', 'PUT'):
            suffix = 'Unsafe'
        elif method == 'GET' and '{uid}' not in path:
            suffix = 'List'
        else:
            suffix = 'Detail'
        try:
            return f'{suffix}{serializer.Meta.ref_name}'
        except AttributeError:
            try:
                name = serializer.__name__.replace(
                    'ModelSerializer', ''
                ).replace('Serializer', '')
            except Exception:
                name = serializer.__class__.__name__.replace(
                    'ModelSerializer', ''
                ).replace('Serializer', '')
            return f'{suffix}{name}'



# coding: utf-8
import logging
import os

from django.conf import settings
from django.dispatch import receiver
from django.db.models.signals import pre_save, post_save

from concrete_datastore.concrete.models import Email  # pylint:disable=E0611
from concrete_datastore.concrete.automation.tasks import (
    send_async_mails,
    perform_send_email,
)


logger = logging.getLogger(__name__)


@receiver(post_save, sender=Email)
def send_email_post_save(sender, instance, **kwargs):
    #: If SMTP_MAILING_USE_ASYNC is enabled, the async task should be in a
    #: post_save signal, because the instance should exist
    if (
        instance.resource_status == 'to-send'
        and settings.SMTP_MAILING_USE_ASYNC is True
    ):
        send_async_mails.apply_async(
            queue='email_senders', kwargs={"email_pk": str(instance.pk)}
        )


@receiver(pre_save, sender=Email)
def send_email_pre_save(sender, instance, **kwargs):
    if (
        instance.resource_status == 'to-send'
        and settings.SMTP_MAILING_USE_ASYNC is False
    ):
        perform_send_email(
            instance=instance,
            enable_retrying=settings.RETRY_ON_SENDING_SYNC_EMAILS,
        )


def build_absolute_uri(uri):
    root = '{}://{}:{}/'.format(
        settings.SCHEME, settings.HOSTNAME, settings.PORT
    )
    if uri.startswith('/') and len(uri) > 1:
        uri = uri[1:]
    return os.path.join(root, uri)



# coding: utf-8
from django.conf import settings

from rest_framework.exceptions import APIException
from rest_framework import permissions, status


class PreconditionFailed(APIException):
    status_code = status.HTTP_412_PRECONDITION_FAILED
    default_detail = 'Precondition Fail to perform this action'
    default_code = 'precondition_failed'


CONCRETE_SETTINGS = getattr(settings, 'CONCRETE', {})
API_PERMISSIONS_CLASSES = CONCRETE_SETTINGS.get('API_PERMISSIONS_CLASSES', {})


def check_minimum_level(method, user, model):
    authenticated = user.is_authenticated
    superuser = user.is_superuser is True
    at_least_admin = False if user.is_anonymous else user.is_at_least_admin
    staff = False if user.is_anonymous else user.is_at_least_staff
    at_least_anonymous = True  #: Always true

    # Set minimum level to do operations on ConcreteRole models
    METHOD_TO_MINIMUM_LEVEL_NAME = {
        'GET': 'authenticated',
        'POST': 'admin',
        'PUT': 'admin',
        'PATCH': 'admin',
        'DELETE': 'admin',
    }

    PERMISSION_LEVELS = {
        'superuser': superuser,
        'admin': at_least_admin,
        'staff': staff,
        'manager': staff,
        'authenticated': authenticated,
        'anonymous': at_least_anonymous,
    }

    level_value = METHOD_TO_MINIMUM_LEVEL_NAME[method]
    user_allowed = PERMISSION_LEVELS[level_value]
    return user_allowed


class BlockedUsersPermission(permissions.BasePermission):
    def has_permission(self, request, view):
        user = request.user
        if not user.is_anonymous:
            if not user.is_confirmed():
                raise PreconditionFailed(
                    detail={
                        'message': 'Email has not been validated',
                        "_errors": ["EMAIL_NOT_VALIDATED"],
                    }
                )
        if request.method not in ["OPTIONS", "HEAD"]:
            if user.is_anonymous:
                return False
            return user.is_at_least_admin
        return True


class ConcreteRolesPermission(permissions.BasePermission):
    message = 'User access permission refused.'

    # Permissions linked to ConcreteRole and ConcretePermission views
    def has_permission(self, request, view):
        if not request.user.is_anonymous:
            if not request.user.is_confirmed():
                raise PreconditionFailed(
                    detail={
                        'message': 'Email has not been validated',
                        "_errors": ["EMAIL_NOT_VALIDATED"],
                    }
                )
        if request.method not in ["OPTIONS", "HEAD"]:
            model = view.model_class
            level_allowed = check_minimum_level(
                request.method, request.user, model
            )
            return level_allowed
        return True



# coding: utf-8
import os
import logging
from django.conf import settings
from django.dispatch import receiver
from django.db.models.signals import pre_delete, post_save
from django.contrib.auth import get_user_model

import concrete_datastore.concrete.models
from concrete_datastore.concrete.models import DIVIDER_MODEL
from concrete_datastore.api.v1.views import (
    remove_instances_user_tracked_fields,
)


logger = logging.getLogger(__name__)


@receiver(pre_delete)
def on_pre_delete(sender, instance, **kwargs):
    model_name = instance.__class__.__name__
    if (
        hasattr(concrete_datastore.concrete.models, model_name)
        and instance.__class__.__name__
        not in settings.IGNORED_MODELS_ON_DELETE
    ):
        # pylint: disable=no-member
        concrete_datastore.concrete.models.DeletedModel.objects.create(
            model_name=model_name, uid=instance.uid
        )

        # Remove files of a deleted instance, if this instance has a FileField
        file_fields = [
            field.name
            for field in instance._meta.fields
            if field.get_internal_type() in ['FileField', 'ImageField']
        ]
        for field_name in file_fields:
            field_value = getattr(instance, field_name)

            if not field_value:
                continue
            try:
                file_path = field_value.path
            except Exception as e:
                logger.exception(
                    f'Cannot retrieve the file path for {field_value} '
                    f'(field {field_name} of instance '
                    f'<{model_name}:{instance.uid}>). The error is {e}'
                )
                continue
            if os.path.exists(file_path) is False:
                continue
            try:
                os.remove(file_path)
            except Exception:
                logger.exception(
                    f'Error while attempting to delete file {file_path} '
                    f'(field {field_name} of instance '
                    f'<{model_name}:{instance.uid}>'
                )
                continue


@receiver(post_save, sender=get_user_model())
def on_post_save(sender, instance, **kwargs):
    if instance.level == 'blocked':
        divider_manager = getattr(
            instance, '{}s'.format(DIVIDER_MODEL.lower())
        )
        user_dividers = divider_manager.values_list('uid', flat=True)
        remove_instances_user_tracked_fields(instance, user_dividers)
        divider_manager.clear()
        instance.concrete_groups.clear()



# coding: utf-8
from collections import OrderedDict
from importlib import import_module

from django.contrib.auth import get_user_model
from django.conf import settings
from django.apps import apps

from rest_framework.reverse import reverse
from rest_framework import serializers
from drf_extra_fields.geo_fields import PointField

from concrete_datastore.concrete.models import LIST_USER_LEVEL
from concrete_datastore.concrete.models import (
    AuthToken,
    get_common_fields,
    get_user_tracked_fields,
    DIVIDER_MODEL,
    TemporaryToken,
)
from concrete_datastore.api.v1.validators import (
    validate_file,
    get_field_validator,
    is_field_required,
)
from concrete_datastore.concrete.constants import MFA_OTP, MFA_EMAIL
from concrete_datastore.concrete.meta import meta_registered
from concrete_datastore.authentication.mfa import is_mfa_enabled
from concrete_datastore.concrete.meta import get_meta_definition_by_model_name
from concrete_datastore.api.v1 import DEFAULT_API_NAMESPACE
from concrete_datastore.api.v1.exceptions import (
    PasswordInsecureValidationError,
)
from concrete_datastore.api.v1.signals import build_absolute_uri


concrete = apps.get_app_config('concrete')


def default_session_control_exempt_rule(user):
    return False


def get_session_control_exempt_rule():
    module_name, func_name = settings.SESSION_CONTROL_DEFAULT_RULE.rsplit(
        '.', 1
    )
    module = import_module(module_name)
    session_control_exempt_rule = getattr(module, func_name)
    return session_control_exempt_rule


class AuthLoginSerializer(serializers.Serializer):
    email = serializers.EmailField()
    password = serializers.CharField()


class ChangePasswordSerializer(serializers.Serializer):
    email = serializers.EmailField()
    password1 = serializers.CharField()
    password2 = serializers.CharField()
    password_change_token = serializers.UUIDField(required=False)


class SecureLoginSerializer(serializers.Serializer):
    token = serializers.UUIDField()


class SecureLoginCodeSerializer(serializers.Serializer):
    email = serializers.EmailField()
    code = serializers.CharField()


class RegisterSerializer(serializers.Serializer):
    email = serializers.EmailField()
    password1 = serializers.CharField(required=False, allow_null=True)
    password2 = serializers.CharField(required=False, allow_null=True)
    email_format = serializers.CharField(required=False, allow_null=True)
    url_format = serializers.CharField(
        required=False,
        allow_null=True,
        default=settings.DEFAULT_REGISTER_URL_FORMAT,
    )

    class Meta:
        fields = (
            "email",
            "password1",
            "password2",
            "email_format",
            "url_format",
        )

    def validate_url_format(self, value):
        if value is None:
            return settings.DEFAULT_REGISTER_URL_FORMAT
        return value


class ResetPasswordSerializer(serializers.Serializer):
    email = serializers.EmailField()
    url_format = serializers.CharField(
        required=False,
        allow_null=True,
        default=settings.DEFAULT_RESET_PASSWORD_URL_FORMAT,
    )

    class Meta:
        fields = ("email", "url_format")

    def validate_url_format(self, value):
        if value is None:
            return settings.DEFAULT_RESET_PASSWORD_URL_FORMAT
        return value


class RetrieveSecureConnectCodeSerializer(serializers.Serializer):
    email = serializers.EmailField()


class UserSerializer(serializers.ModelSerializer):
    url = serializers.SerializerMethodField()
    token = serializers.SerializerMethodField()
    is_verified = serializers.SerializerMethodField()
    mfa_mode = serializers.SerializerMethodField()
    password = serializers.CharField(write_only=True, required=False)
    email = serializers.EmailField(required=True)
    level = serializers.CharField()
    groups = serializers.SerializerMethodField()

    api_namespace = DEFAULT_API_NAMESPACE

    def __init__(self, api_namespace, *args, **kwargs):
        self.api_namespace = api_namespace
        super(UserSerializer, self).__init__(*args, **kwargs)

    class Meta:
        model = get_user_model()
        fields = (
            "uid",
            'email',
            "url",
            "token",
            'first_name',
            "password",
            "last_name",
            "level",
            'is_verified',
            'groups',
            'external_auth',
            'mfa_mode',
            "{}s".format(DIVIDER_MODEL.lower()),
        )

    def get_groups(self, user):
        return user.concrete_groups.values_list('name', flat=True)

    def _get_token_key(self, user):
        if settings.USE_MULTIPLE_TOKENS is False:
            token = AuthToken.objects.filter(user=user, expired=False).first()
            if token:
                return token.key
        token = AuthToken.objects.create(user=user)
        return token.key

    def get_url(self, obj):
        uri = reverse("{}:account-me".format(self.api_namespace))
        return build_absolute_uri(uri)

    def get_is_verified(self, obj):
        if is_mfa_enabled(user=obj) is False:
            return True

        return self.context.get('is_verified', False)

    def get_mfa_mode(self, obj):
        if obj.totp_device is not None:
            return MFA_OTP
        if is_mfa_enabled(user=obj) is True:
            return MFA_EMAIL
        return ''

    def manage_session_token(self, obj):
        key = self._get_token_key(user=obj)

        rule = get_session_control_exempt_rule()
        user_is_session_control_exempted = rule(user=obj)

        session_control_enabled = settings.SESSIONS_NUMBER_CONTROL_ENABLED
        max_sessions_allowed = settings.MAX_SIMULTANEOUS_SESSIONS
        unlimited_sessions = settings.MAX_SIMULTANEOUS_SESSIONS == 0
        user_active_tokens = (
            AuthToken.objects.filter(user=obj, expired=False)
            .exclude(key=key)
            .order_by('expiration_date')
        )
        if (
            session_control_enabled
            and user_active_tokens.count() >= max_sessions_allowed
            and not unlimited_sessions
            and not user_is_session_control_exempted
        ):
            # We expire the most ancient tokens, while keeping the number of
            # active tokens >= MAX_SIMULTANEOUS_SESSIONS
            # NB: the recently created token is not included in the queryset
            # that's why we add the +1 to the tokens_count
            tokens_count = user_active_tokens.count()
            nb_of_token_to_expire = tokens_count - max_sessions_allowed + 1
            tokens_to_expire_pks = user_active_tokens.values_list(
                'pk', flat=True
            )[:nb_of_token_to_expire]
            AuthToken.objects.filter(pk__in=tokens_to_expire_pks).update(
                expired=True
            )

        return key

    def get_token(self, obj):
        #: If the user has not the minimum level for two factor
        #: Return the final token
        if is_mfa_enabled(user=obj):
            #: With two factor auth enable
            #: return the token if identity has been validated
            is_verified = self.context.get('is_verified', False)
            if is_verified:
                token_key = self.manage_session_token(obj)
            else:
                token = TemporaryToken.objects.create(user=obj)
                token_key = token.key

        else:
            token_key = self.manage_session_token(obj)
        return token_key


def make_related_serializer_class(
    target_model_name, many, nested=False, api_namespace=DEFAULT_API_NAMESPACE
):
    if target_model_name not in meta_registered:
        raise ValueError(f'Related to unknown model {target_model_name}')

    rel_meta_model = meta_registered[target_model_name]
    rel_serializer_class = make_serializer_class(
        meta_model=rel_meta_model, nested=nested, api_namespace=api_namespace
    )

    return rel_serializer_class(many=many)


def make_custom_serializer_fields(
    meta_model, api_namespace=DEFAULT_API_NAMESPACE
):
    custom_fields_attrs = {}

    CONCRETE_SETTINGS = getattr(settings, 'CONCRETE', {})
    SERIALIZERS_SETTINGS = CONCRETE_SETTINGS.get('SERIALIZERS', {})

    for model_name, SERIALIZER_SETTINGS in SERIALIZERS_SETTINGS.items():
        if meta_model.get_model_name() == model_name:
            CUSTOM_FIELDS = SERIALIZER_SETTINGS.get('CUSTOM_FIELDS', {})

            for field_name, field_args in CUSTOM_FIELDS.items():
                if 'type' not in field_args:
                    raise ValueError(
                        'CONCRETE improperly configured : custom field '
                        f'{field_name} should have a type attribute'
                    )

                if field_args['type'] == 'RelatedModelSerializer':
                    if 'to' not in field_args:
                        raise ValueError(
                            'CONCRETE improperly configured : custom field '
                            f'{field_name} with type RelatedModelSerializer '
                            'should have a "to" argument'
                        )

                    custom_fields_attrs.update(
                        {
                            field_name: make_related_serializer_class(
                                target_model_name=field_args['to'],
                                many=True,
                                nested=False,
                                api_namespace=api_namespace,
                            )
                        }
                    )
                else:
                    raise NotImplementedError(
                        'Generic type for custom field in serializer is '
                        'not implemented in this version of concrete'
                    )

    return custom_fields_attrs.keys(), custom_fields_attrs


def get_field_extra_kwargs(field):
    extra_kwargs = {'required': is_field_required(field)}
    #: All char and text fields must be trimmed, except the choice fields.
    #: Choice field is a DRF model serializer field is generated automatically
    #: when encountering a Django CharField having a "choices" argument.
    #: ChoiceFields do not accept the argument "trim_whitespace".
    is_char = field.f_type in ('CharField', 'TextField')
    is_choice_field = 'choices' in field.f_args
    if is_char and not is_choice_field:
        extra_kwargs.update(trim_whitespace=False)
    return extra_kwargs


def make_serializer_class(
    meta_model,
    api_namespace=DEFAULT_API_NAMESPACE,
    limit_fields=None,
    nested=True,
    safe=False,
):
    enum_fields = list(meta_model.get_fields())
    dict_fields = OrderedDict(enum_fields)
    _all_fields = (
        list(dict_fields.keys())
        + list(get_common_fields().keys())
        + ['url', 'verbose_name']
    )

    custom_fields_names, custom_fields_attrs = make_custom_serializer_fields(
        meta_model, api_namespace=api_namespace
    )

    _all_fields += custom_fields_names
    if meta_model.get_model_name() not in ('User',):
        _all_fields += get_user_tracked_fields().keys()
    if meta_model.get_model_name() not in ('User', DIVIDER_MODEL):
        _all_fields += ['scopes']
    api_display_fields = meta_model.get_property('m_api_list_display') or []
    if api_display_fields == []:
        if limit_fields is None:
            _fields = _all_fields
        else:
            _fields = [f for f in _all_fields if f in limit_fields]
    else:
        _fields = api_display_fields
    if meta_model.get_model_name() == 'User' and api_display_fields == []:
        _fields += [
            'last_name',
            'first_name',
            "{}s".format(DIVIDER_MODEL.lower()),
            "admin",
            "is_staff",
            "level",
            "unsubscribe_notification_url",
            "unsubscribe_all",
            "unsubscribe_to",
            "external_auth",
        ]

        if safe:
            _fields += ['email']

    # TODO: rajouter les <name>_uid dans _fields
    fk_read_only_fields = []
    for name, field in enum_fields:
        if field.type.startswith("rel_"):
            _fields += ['{}_uid'.format(name)]
            fk_read_only_fields += [name]

    class Meta:
        model = concrete.models[meta_model.get_model_name().lower()]
        fields = _fields

        read_only_fields = (
            ['uid', 'created_by', 'admin', 'is_staff']
            + fk_read_only_fields
            + [f for f in _all_fields if f.startswith('resource_')]
        )
        extra_kwargs = {
            name: get_field_extra_kwargs(field) for name, field in enum_fields
        }

        # TODO: DEACTIVATED by LCO on 06/11/18
        # vars_model = vars(model)
        # for key, value in vars_model.items():
        #     # TODO: Exclude model User
        #     if isinstance(value, ReverseManyToOneDescriptor):
        #         if not (key.startswith('divider') or (key.startswith('can'))):
        #             fields.append(key)
        #             read_only_fields.append(key)

    attrs = {'Meta': Meta}

    # TODO : if field is relational, expose pk and url serialized
    for name, field in enum_fields:
        if field.f_type == 'FileField':
            attrs.update(
                {
                    name: serializers.FileField(
                        required=False,
                        allow_null=True,
                        validators=[validate_file],
                    )
                }
            )
        if field.f_type == 'PointField':
            attrs.update(
                {
                    name: PointField(
                        required=not field.f_args.get("blank", False),
                        allow_null=True,
                    )
                }
            )
        if field.f_type == 'JSONField':
            attrs.update(
                {name: serializers.JSONField(binary=False, required=False)}
            )
        if field.type.startswith("rel_") and nested is True:
            force_nested = getattr(field, 'force_nested', False)

            attrs.update(
                {
                    name: make_related_serializer_class(
                        target_model_name=field.f_args['to'],
                        many=(field.type == 'rel_iterable'),
                        nested=force_nested,
                        api_namespace=api_namespace,
                    )
                }
            )

    attrs.update(custom_fields_attrs)

    class _ModelSerializer(serializers.ModelSerializer):
        url = serializers.SerializerMethodField()
        verbose_name = serializers.SerializerMethodField()
        scopes = serializers.SerializerMethodField()

        def get_scopes(self, obj):
            divider = getattr(obj, DIVIDER_MODEL.lower(), None)
            if divider:
                return {"entity_uid": divider.uid}
            return None

        def get_url(self, obj):
            uri = reverse(
                "{}:{}-detail".format(
                    api_namespace, meta_model.get_dashed_case_class_name()
                ),
                args=(obj.pk,),
            )
            if hasattr(self, '_context'):
                if 'request' in self._context:
                    request = self._context['request']
                    return request.build_absolute_uri(uri)
            return build_absolute_uri(uri)  # skip-test-coverage

        def get_verbose_name(self, obj):
            try:
                return str(obj)
            except Exception:
                return ''

    for name, field in enum_fields:
        attrs.update({f'validate_{name}': get_field_validator(field=field)})
        if field.type.startswith("rel_i"):
            x = field.f_args["to"].split('.')
            field_model = apps.get_model(app_label=x[0], model_name=x[1])
            required = not field.f_args.get("blank", False)
            attrs.update(
                {
                    '{}_uid'.format(name): serializers.PrimaryKeyRelatedField(
                        source=name,
                        many=True,
                        allow_null=True,
                        required=required,
                        queryset=field_model.objects.all(),
                    )
                }
            )

        if field.type.startswith("rel_s"):
            # Ex: concrete.Category, split to get
            # Django model with apps.get_model()
            x = field.f_args["to"].split('.')
            field_model = apps.get_model(app_label=x[0], model_name=x[1])
            required = not field.f_args.get("blank", False)
            null = field.f_args.get("null", False)
            attrs.update(
                {
                    '{}_uid'.format(name): serializers.PrimaryKeyRelatedField(
                        many=False,
                        source=name,
                        allow_null=null,
                        required=required,
                        queryset=field_model.objects.all(),
                    )
                }
            )

    if meta_model.get_model_name() == 'User':
        attrs.update({'level': serializers.CharField()})

        def validate_level(self, value):
            value = value.lower()

            if value not in LIST_USER_LEVEL:
                raise serializers.ValidationError("Wrong value")

            else:
                user_level = self.context.get('request').user.level
                forbidden_for_admin = user_level == "admin" and value in [
                    "admin",
                    "superuser",
                ]
                forbidden_for_manager = user_level == "manager" and value in [
                    "admin",
                    "superuser",
                    "manager",
                    "blocked",
                ]
                forbidden_for_simple = user_level == "simpleuser"
                if (
                    forbidden_for_manager
                    or forbidden_for_admin
                    or forbidden_for_simple
                ):
                    raise serializers.ValidationError(
                        "You don't have permission to set this level"
                    )
            return value

        _ModelSerializer.validate_level = validate_level

        attrs.update(
            {
                'unsubscribe_notification_url': serializers.SerializerMethodField(
                    'get_unsubscribe_url'
                )
            }
        )

        def get_unsubscribe_url(self, obj):
            return '{scheme}://{domain}:{port}{uri}'.format(
                scheme=settings.SCHEME,
                domain=settings.HOSTNAME,
                port=settings.PORT,
                uri=reverse(
                    'concrete:unsubscribe_notifications',
                    kwargs={'token': obj.subscription_notification_token},
                ),
            )

        _ModelSerializer.get_unsubscribe_url = get_unsubscribe_url

    api_model_serializer = type(
        str('{}ModelSerializer'.format(meta_model.get_model_name())),
        (_ModelSerializer,),
        attrs,
    )
    return api_model_serializer


def make_account_me_serialier(api_namespace=DEFAULT_API_NAMESPACE):
    meta = get_meta_definition_by_model_name('User')
    meta_serializer_class = make_serializer_class(
        meta_model=meta, api_namespace=api_namespace, nested=False, safe=True
    )

    Meta = meta_serializer_class.Meta

    Meta.fields = [
        field
        for field in Meta.fields
        if field not in ['admin', 'is_staff', 'is_superuser', 'is_active']
    ] + ['level']

    Meta.read_only_fields = Meta.read_only_fields + [
        'uid',
        'public',
        'creation_date',
        'modification_date',
        '{}s'.format(DIVIDER_MODEL.lower()),
    ]

    return type(
        str('AccountMeSerializer'),
        (meta_serializer_class,),
        {
            'Meta': Meta,
            'level': serializers.SerializerMethodField(),
            'get_level': lambda self, obj: obj.get_level(),
        },
    )


class ConcretePasswordValidator(object):
    def __call__(self, value):
        try:
            for validator in settings.AUTH_PASSWORD_VALIDATORS:
                module_name, validator_name = validator['NAME'].rsplit('.', 1)
                module = import_module(module_name)
                validator_cls = getattr(module, validator_name)
                validator_cls().validate(value)
        except PasswordInsecureValidationError:
            raise
        except Exception as e:
            raise PasswordInsecureValidationError(
                str(e), code='PASSWORD_INSECURE'
            )



# coding: utf-8
import random
import string

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError


def generate_random_password(
    length=24, chars=string.ascii_letters + string.digits
):
    return ''.join(random.choice(chars) for _ in range(length))  # nosec


def get_admin_url():
    port = ''
    if int(settings.PORT) not in (80, 443):
        port = ':{}'.format(settings.PORT)

    admin_url = '{}://{}{}/concrete-datastore-admin/'.format(
        settings.SCHEME, settings.HOSTNAME, port
    )
    return admin_url


class Command(BaseCommand):
    help = 'Reset password'

    def add_arguments(self, parser):
        parser.add_argument('email', type=str, help='user email')

    def handle(self, *args, **options):
        concrete = apps.get_app_config('concrete')
        EmailModel = concrete.models['email']
        UserModel = get_user_model()
        email = options['email'].lower()

        try:
            user = UserModel.objects.get(email=email)
        except UserModel.DoesNotExist:
            raise CommandError(f"This email: {email} does not exist")

        password = generate_random_password()
        user.set_password(password)
        user.save()

        admin_url = get_admin_url()
        email_instance = EmailModel(
            created_by=user,
            subject='Reset password to Concrete Instance',
            body=settings.RESET_PASSWORD_EMAIL_MESSAGE_BODY.format(
                hostname=settings.HOSTNAME,
                admin_url=admin_url,
                email=email,
                password=password,
            ),
            receiver=user,
        )

        email_instance.save()



# conding: utf-8
from django import forms
from django.core.exceptions import ValidationError
from django.contrib.auth import get_user_model
from django.utils.translation import ngettext_lazy, gettext_lazy as _
from concrete_datastore.concrete.constants import MFA_OTP, MFA_EMAIL


class ConfigureOTPLoginForm(forms.Form):
    otp_error_messages = {
        'token_required': _('Please enter your OTP token.'),
        'challenge_exception': _('Error generating challenge: {0}'),
        'challenge_message_email': _(
            'Enter the two authentication code you received by email'
        ),
        'invalid_token': _(
            'Invalid token. Please make sure you have entered it correctly.'
        ),
        'n_failed_attempts': ngettext_lazy(
            "Verification temporarily disabled because of %(failure_count)d failed attempt, please try again soon.",
            "Verification temporarily disabled because of %(failure_count)d failed attempts, please try again soon.",
            "failure_count",
        ),
        'verification_not_allowed': _(
            "Verification of the token is currently disabled"
        ),
        'otp_already_configured': _('OTP authentication already configured'),
        'wrong_auth_credentials': _('Wrong auth credentials'),
    }
    email = forms.EmailField(label="Email", max_length=254)
    password = forms.CharField(label="Password", max_length=254)

    otp_token = forms.CharField(
        required=False, widget=forms.TextInput(attrs={'autocomplete': 'off'})
    )

    def __init__(self, *args, **kwargs):
        self.user_cache = None
        super().__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super().clean()
        email = cleaned_data.get("email")
        password = cleaned_data.get("password")
        otp_token = cleaned_data.get("otp_token")
        User = get_user_model()
        self.user_cache = User.objects.filter(email=email).first()
        if self.user_cache is None:
            raise ValidationError(
                self.otp_error_messages['wrong_auth_credentials'],
                code='wrong_auth_credentials',
            )
        if self.user_cache.check_password(password) is False:
            #: Clear user cache to reset the form
            self.user_cache = None
            raise ValidationError(
                self.otp_error_messages['wrong_auth_credentials'],
                code='wrong_auth_credentials',
            )
        otp_device = self.user_cache.emaildevice_set.filter(
            mfa_mode=MFA_OTP, confirmed=True
        ).first()
        if otp_device:
            #: Clear user cache to reset the form
            self.user_cache = None
            raise ValidationError(
                self.otp_error_messages['otp_already_configured'],
                code='otp_already_configured',
            )
        device = self._chosen_device(self.user_cache)
        if not otp_token:
            self._handle_challenge(device)
        else:
            verified = device._verify_email_token(otp_token)
            if verified is False:
                #: Keep user cache to avoid selecting an email and password
                raise ValidationError(
                    self.otp_error_messages['invalid_token'],
                    code='invalid_token',
                )
        return cleaned_data

    def _handle_challenge(self, device):
        try:
            device.generate_challenge()
        except Exception as e:
            raise forms.ValidationError(
                self.otp_error_messages['challenge_exception'].format(e),
                code='challenge_exception',
            )

        #:  If the challenge is successfully generated and in order to show
        #:  the appropriate message, a `ValidationError` must be raised.
        #:  *NB* This validation error does not mean that an error occured
        raise forms.ValidationError(
            self.otp_error_messages['challenge_message_email'],
            code='challenge_message_email',
        )

    def _chosen_device(self, user):
        device = user.emaildevice_set.filter(
            confirmed=True, mfa_mode=MFA_EMAIL
        ).first()
        if not device:
            device = user.emaildevice_set.create(
                email=user.email, name='User default email', confirmed=True
            )

        return device

    def get_user(self):
        return self.user_cache



#: coding: utf-8
from django.conf import settings
from django.core.management.base import BaseCommand
from concrete_datastore.concrete import models as concrete_models


def get_sorted_stats(key):
    data_model = settings.META_MODEL_DEFINITIONS
    models = data_model['manifest']['data_modeling']['models']
    model_names = map(lambda x: x['name'], models)
    models_count = [
        {'name': name, 'count': getattr(concrete_models, name).objects.count()}
        for name in model_names
    ]
    reverse = key.startswith('-')
    return sorted(
        models_count, key=lambda x: x[key.replace('-', '')], reverse=reverse
    )


class Command(BaseCommand):
    help = 'Displays a list of all models and the number of instances in DB'

    def add_arguments(self, parser):
        parser.add_argument(
            '--key',
            type=str,
            help=(
                'The key used to sort the display: * "name": the result will '
                'be sorted by ascending based on the model name * "-name": '
                'the result will be sorted by descending based on the model '
                'name * "count": the result will be sorted by ascending based '
                'on the number of instances * "-count": the result will be '
                'sorted by descending based on the number of instances'
            ),
            choices=('name', 'count', '-name', '-count'),
            default='-count',
        )

    def handle(self, *args, **options):
        key = options['key']
        for elt in get_sorted_stats(key=key):
            print(f'{elt["name"]}: {elt["count"]:_} objetcs')



# coding: utf-8
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Create users'

    def add_arguments(self, parser):
        parser.add_argument('email', type=str, help='choosen email')
        parser.add_argument('password', type=str, help='choosen password')

    def handle(self, *args, **options):
        UserModel = get_user_model()
        email = options['email'].lower()

        user, created = UserModel.objects.get_or_create(email=email)

        if created:
            user.set_level('superuser')
            if options['password'] is not None:
                user.set_password(options['password'])

            user.save()



# coding: utf-8
from concrete_datastore.parsers.models import Model_V1
from concrete_datastore.parsers.meta import (
    make_model_cls,
    make_modelisation_cls,
)


def loads_models_v0(model_definitions):
    meta_models = []
    for meta_model_definition in model_definitions:
        meta_model_definition_copy = meta_model_definition.copy()
        meta_models += [make_model_cls(meta_model_definition_copy)]

    return meta_models


def loads_models_v1(model_definitions):
    try:
        modelisation = make_modelisation_cls(
            model_definitions, version='1.0.0', base=Model_V1
        )()
    except KeyError:
        raise KeyError('Invalid meta model definition')
    return modelisation.get_meta_models()  # pylint: disable=no-member


data_models_version = {"0.0.0": loads_models_v0, "1.0.0": loads_models_v1}


def loads_meta(model_definitions):
    if isinstance(model_definitions, (list, tuple)):
        return data_models_version["0.0.0"](model_definitions)
    try:
        version = model_definitions['manifest']['version']
    except KeyError:
        raise KeyError('Invalid meta model definition')

    load_func = data_models_version[version]
    return load_func(model_definitions)



# coding: utf-8
import uuid
import logging
from collections import namedtuple, defaultdict
from copy import deepcopy
from keyword import iskeyword
from six import with_metaclass
from django.core.validators import validate_ipv46_address
from concrete_datastore.parsers.exceptions import (
    UnknownDatamodelVersionError,
    MissingRelationForModel,
    DuplicatedRelationForModel,
    MissingPermissionsOrQueriesForModel,
    DuplicatedPermissionsOrQueriesForModel,
    NameNotAllowed,
    DuplicatedFieldsError,
    DuplicatedReverseForModel,
    UnknownDatatypeForField,
    MissingKeyForDefinition,
    ProtectedModelNameError,
    ProtectedFieldNameError,
    UnknownIPProtocol,
)
from concrete_datastore.parsers.validators import validate_specifier
from concrete_datastore.parsers.constants import (
    VERSIONS_ATTRIBUTES,
    CONCRETE_USER_PROTECTED_FIELDS,
    CONCRETE_MODELS_PROTECTED_FIELDS,
    CONCRETE_CUSTOM_MODELS,
    PROTOCOL_EQUIVALENCE,
    AUTHORIZED_IP_PROTOCOLS,
)
from concrete_datastore.parsers.models import Model
from concrete_datastore.parsers.fields import (
    field_descriptors,
    get_field_descriptor,
)
from concrete_datastore.parsers.helpers import replace_prefix, remove_prefix


class DynamicMetaClass(type):
    def __new__(cls, name, bases, clsdict):
        msg = 'Creating new Dynamic Class {}'.format(name)
        logging.debug(msg)

        if 'specifier' not in clsdict:
            raise RuntimeError('Missing specifier in class {}'.format(name))

        specifier_dict = clsdict['specifier']
        validate_specifier(specifier_dict=specifier_dict)

        dyn_class_name = specifier_dict['std.name']
        attr_names = list(map(replace_prefix, specifier_dict.keys()))
        attr_names.extend(map(remove_prefix, specifier_dict.keys()))
        struct_name = dyn_class_name.title()
        struct = namedtuple(struct_name, attr_names)
        struct_kwargs = {
            replace_prefix(k): v for k, v in specifier_dict.items()
        }
        struct_kwargs.update(
            {remove_prefix(k): v for k, v in specifier_dict.items()}
        )
        clsdict['_specifier'] = struct(**struct_kwargs)
        clsdict.pop('specifier')

        msg = 'Created {} with structure {}'.format(
            name, clsdict['_specifier']
        )
        logging.debug(msg)

        return type.__new__(cls, name, bases, clsdict)


def make_cls(spec, base, prefix):
    dyn_class_name = spec['std.name']

    def __init__(self, *args, **kwargs):
        self._uid = uuid.uuid4()
        super(self.__class__, self).__init__(*args, **kwargs)

    def __hash__(self):
        return self._uid.int

    return type(
        dyn_class_name.title(),
        (with_metaclass(DynamicMetaClass, base),),
        {'specifier': spec, '__init__': __init__, '__hash__': __hash__},
    )


def make_field_cls(spec, base=object):
    return make_cls(spec, base, prefix='Field')


def make_model_cls(spec, base=Model):
    fields = []
    for field_spec in spec['std.fields']:
        validate_specifier(field_spec)
        field_type = field_spec['std.type']
        field_cls = make_field_cls(
            field_spec, base=field_descriptors[field_type]
        )
        field_instance = field_cls(name=field_cls._specifier.std_name)
        fields.append((field_cls._specifier.std_name, field_instance))

    spec['std.fields'] = [f[1]._specifier for f in fields]
    model_cls = make_cls(spec, base, prefix='')

    # From 0.3.0 (incompatible) :
    # model_cls = make_cls(spec, base, prefix='Model')

    for name, field in fields:
        setattr(model_cls, name, field)
    return model_cls


def make_modelisation_cls(modelisation_spec, version, base=Model):
    spec = deepcopy(modelisation_spec)

    manifest = spec['manifest']
    data_modeling = manifest['data_modeling']

    models = data_modeling['models']

    many_to_many_relations = data_modeling['many_to_many_relations']
    one_to_many_relations = data_modeling['one_to_many_relations']
    permissions = data_modeling['permissions']
    resource_queries = data_modeling['resource_queries']

    relations = {"m2m": many_to_many_relations, "fk": one_to_many_relations}
    attrs = VERSIONS_ATTRIBUTES[version]

    def get_permissions(permission_dict, version):
        if version != '1.0.0':
            raise UnknownDatamodelVersionError()
        minimum_levels = {}
        for perm, level in permission_dict['minimum_levels'].items():
            key = 'creation' if perm == 'create' else perm
            minimum_levels.update({f'm_{key}_minimum_level': level})
        return minimum_levels

    def get_resource_queries(resource_dict, version):
        if version != '1.0.0':
            raise UnknownDatamodelVersionError()
            # Ordering fields default value is display_fields to stay
            # retrocompatible
        return {
            'm_search_fields': resource_dict['search_fields'],
            'm_filter_fields': resource_dict['filter_fields'],
            'm_ordering_fields': resource_dict.get(
                'ordering_fields', resource_dict['display_fields']
            )
            + ['creation_date', 'modification_date'],
            'm_distance_filter_field': resource_dict.get(
                'distance_filter_field', ''
            ),
            'm_export_fields': resource_dict['export_fields'],
            'm_list_display': resource_dict['display_fields']
            + ['creation_date', 'modification_date'],
        }

    def get_relation_for_field(
        self,
        f_type,
        src_field,
        src_model_uid,
        src_model_name,
        target_model_name,
        target_model_uid,
    ):
        if f_type == 'fk':
            rel_type = 'one_to_many_relations'
        elif f_type == 'm2m':
            rel_type = 'many_to_many_relations'
        relation_elements = list(
            filter(
                lambda x: x['source_field'] == src_field
                and x['source_model'][self.element_id] == src_model_uid
                and x['target_model'][self.element_id] == target_model_uid,
                relations[f_type],
            )
        )
        if len(relation_elements) == 0:
            raise MissingRelationForModel(
                rel_type=rel_type,
                source=src_model_name,
                target=target_model_name,
                field_name=src_field,
            )
        if len(relation_elements) > 1:
            raise DuplicatedRelationForModel(
                rel_type=rel_type,
                source=src_model_name,
                target=target_model_name,
                field_name=src_field,
            )
        rel = relation_elements[0]
        rel_model_name = rel['target_model']['name']
        return (f'concrete.{rel_model_name}', rel.get('onDelete', 'PROTECT'))

    def get_meta_models(self):
        meta_models = []
        self.models_attrs = defaultdict(lambda: defaultdict(list))
        for meta_model_definition in models:
            self.validate_specifier(meta_model_definition)
            model_name = meta_model_definition[self.element_name]
            if model_name in CONCRETE_CUSTOM_MODELS:
                raise ProtectedModelNameError(model_name=model_name)
            if iskeyword(model_name):
                raise NameNotAllowed(name=model_name, resource_type='model')
            model_permissions = self.get_parameter_for_model(
                model_spec=meta_model_definition,
                parameter_dict=permissions,
                param_type='permissions',
            )

            #:  For now, if the modelisation does not contain a
            #:  'resource_queries' field, we set all of search, filter, export
            #:  and list fields to all the fields of the model
            for field in meta_model_definition[self.fields_spec]:
                self.validate_specifier(field, spec_type='Field')
                field_name = field[self.element_name]
                if iskeyword(field_name):
                    raise NameNotAllowed(
                        name=field_name, resource_type='field'
                    )
                protected_user_field = (
                    model_name.lower() == 'user'
                    and field_name in CONCRETE_USER_PROTECTED_FIELDS
                )
                custom_model_protected_field = (
                    model_name.lower() != 'user'
                    and field_name in CONCRETE_MODELS_PROTECTED_FIELDS
                )
                if protected_user_field or custom_model_protected_field:
                    raise ProtectedFieldNameError(
                        field_name=field_name, model_name=model_name
                    )

                if (
                    field_name
                    in self.models_attrs[model_name][self.fields_spec]
                ):
                    raise DuplicatedFieldsError(
                        field_name=field_name, model_name=model_name
                    )
                self.models_attrs[model_name][self.fields_spec].append(
                    field_name
                )
            all_fields = [
                field[self.element_name]
                for field in meta_model_definition[self.fields_spec]
                if field[self.field_type_spec]
                in [
                    'bool',
                    'txt',
                    'text',
                    'char',
                    'url',
                    'float',
                    'datetime',
                    'date',
                    'int',
                    'fk',
                    'ip',
                ]
                #: we should consider only simple types
            ]
            default_resource_queries = {
                'search_fields': all_fields,
                'filter_fields': all_fields,
                'export_fields': all_fields,
                'display_fields': all_fields,
            }
            model_resource_queries = self.get_parameter_for_model(
                model_spec=meta_model_definition,
                parameter_dict=resource_queries,
                param_type='ressource_queries',
                default=default_resource_queries,
            )
            meta_models += [
                self.make_model_cls(
                    meta_model_definition,
                    model_permissions,
                    model_resource_queries,
                )
            ]
        return meta_models

    def make_model_cls(self, spec, permissions, resource_queries):
        fields = []
        for field_spec in spec[self.fields_spec]:
            field_type = field_spec[self.field_type_spec]
            field_cls = self.make_field_cls(
                field_spec=field_spec,
                model_name=spec[self.element_name],
                base=get_field_descriptor(field_type),
                model_uid=spec['uid'],
            )
            field_instance = field_cls(name=field_cls._specifier.name)
            fields.append((field_cls._specifier.name, field_instance))

        spec['fields_spec'] = [f[1]._specifier for f in fields]
        model_cls = self.make_cls(
            spec=spec,
            base=base,
            permissions=permissions,
            resource_queries=resource_queries,
        )

        for name, field in fields:
            setattr(model_cls, name, field)
        return model_cls

    def validate_specifier(self, field_spec, spec_type='Model'):
        specifier_keys = self.std_verifier[spec_type]

        for key in specifier_keys:
            if key not in field_spec:
                raise MissingKeyForDefinition(
                    key=key, keys_list=specifier_keys, resource_type=spec_type
                )

    def make_field_cls(
        self, field_spec, model_name, base=object, model_uid=None
    ):
        return self.make_cls(
            field_spec,
            base,
            specifier='Field',
            model_name=model_name,
            model_uid=model_uid,
        )

    def update_specifier_data(
        self,
        specifier,
        spec,
        model_name,
        model_uid,
        permissions=None,
        resource_queries=None,
    ):
        if specifier == 'Field':
            datatype = spec.pop(self.field_type_spec)
            if datatype not in self.equivalence_table.keys():
                raise UnknownDatatypeForField(
                    field_name=spec[self.element_name],
                    model_name=model_name,
                    datatype=datatype,
                )
            spec.update({'f_type': self.equivalence_table[datatype]})
            attributes = spec.pop('attributes')
            if 'allow_empty' in attributes.keys():
                allow_empty = attributes.pop('allow_empty')
                attributes.update({'null': allow_empty, 'blank': allow_empty})
            if datatype == 'char':
                attributes.setdefault('max_length', 250)
            if datatype == 'ip':
                attributes.setdefault('protocol', 'ipv4_6')
                protocol = attributes['protocol'].lower()
                if protocol not in AUTHORIZED_IP_PROTOCOLS:
                    raise UnknownIPProtocol(
                        protocol, model_name, spec[self.element_name]
                    )
                default_ip = attributes.get('default')
                if default_ip:
                    #: Raise Validation Error if the default IP is invalid
                    validate_ipv46_address(default_ip)
                attributes['protocol'] = PROTOCOL_EQUIVALENCE[protocol]
            if datatype in ['fk', 'm2m']:
                field_type = (
                    f'rel_{"single" if datatype == "fk" else "iterable"}'
                )
                target_field = (
                    attributes.pop('reverse')
                    if 'reverse' in attributes.keys()
                    else None
                )
                target_model_name = attributes['to'][self.element_name]
                if (
                    target_field
                    in self.models_attrs[target_model_name]['reverses']
                ):
                    raise DuplicatedReverseForModel(
                        name=model_name, reverse_name=target_field
                    )
                self.models_attrs[target_model_name]['reverses'].append(
                    target_field
                )
                rel_model, on_delete_rule = self.get_relation_for_field(
                    f_type=datatype,
                    src_field=spec[self.element_name],
                    src_model_uid=model_uid,
                    src_model_name=model_name,
                    target_model_name=target_model_name,
                    target_model_uid=attributes['to'][self.element_id],
                )
                attributes.update({'to': rel_model})
                attributes.update(
                    {'on_delete': on_delete_rule}
                ) if datatype == 'fk' else None
                attributes.update(
                    {'related_name': target_field}
                ) if target_field is not None else None

            else:
                field_type = 'data'

            spec.update({'f_args': attributes})
            spec.update({'type': field_type})

        elif specifier == 'Model':
            spec.update({'name': spec[self.element_name]})
            spec.update({'lookups': permissions['lookups']})
            spec.update(get_resource_queries(resource_queries, self.version))
            spec.update(get_permissions(permissions, self.version))
            spec.update({'m_unique_together': spec.get('unique_together', [])})
            spec.update({'m_unicode': spec.get('representation_field', '')})
            spec.update(
                {'m_is_default_public': spec.get('is_default_public', False)}
            )

    def get_parameter_for_model(
        self, model_spec, param_type, parameter_dict, default=None
    ):
        model_parameter = list(
            filter(
                lambda x: x['model_uid'] == model_spec[self.element_id],
                parameter_dict,
            )
        )
        if len(model_parameter) == 0:
            if default is not None:
                return default
            raise MissingPermissionsOrQueriesForModel(
                param_type=param_type, model_name=model_spec[self.element_name]
            )
        if len(model_parameter) > 1:
            raise DuplicatedPermissionsOrQueriesForModel(
                param_type=param_type, model_name=model_spec[self.element_name]
            )
        return model_parameter[0]

    def make_cls(
        self,
        spec,
        base,
        specifier='Model',
        model_name=None,
        model_uid=None,
        permissions=None,
        resource_queries=None,
    ):
        spec_copy = deepcopy(spec)
        self.update_specifier_data(
            specifier=specifier,
            spec=spec_copy,
            model_name=model_name,
            model_uid=model_uid,
            permissions=permissions,
            resource_queries=resource_queries,
        )

        class DynamicMetaClass(type):
            def __new__(cls, name, bases, clsdict):
                msg = 'Creating new Dynamic Class {}'.format(name)
                logging.debug(msg)

                if 'specifier' not in clsdict:
                    raise RuntimeError(
                        'Missing specifier in class {}'.format(name)
                    )
                specifier_dict = clsdict['specifier']

                dyn_class_name = specifier_dict[self.element_name]
                attr_names = list(specifier_dict.keys())

                struct_name = dyn_class_name
                struct = namedtuple(struct_name, attr_names)
                clsdict['_specifier'] = struct(**specifier_dict)
                clsdict.pop('specifier')

                msg = 'Created {} with structure {}'.format(
                    name, clsdict['_specifier']
                )
                logging.debug(msg)

                return type.__new__(cls, name, bases, clsdict)

        dyn_class_name = spec_copy[self.element_name]

        def __init__(self, *args, **kwargs):
            self._uid = uuid.uuid4()
            super(self.__class__, self).__init__(*args, **kwargs)

        def __hash__(self):
            return self._uid.int

        return type(
            dyn_class_name,
            (with_metaclass(DynamicMetaClass, base),),
            {
                'specifier': spec_copy,
                '__init__': __init__,
                '__hash__': __hash__,
            },
        )

    attrs.update(
        {
            'version': version,
            'get_meta_models': get_meta_models,
            'make_model_cls': make_model_cls,
            'make_field_cls': make_field_cls,
            'make_cls': make_cls,
            'validate_specifier': validate_specifier,
            'update_specifier_data': update_specifier_data,
            'get_parameter_for_model': get_parameter_for_model,
            'get_relation_for_field': get_relation_for_field,
        }
    )

    return type('ModelisationClass_{}'.format(version), (object,), attrs)



# coding: utf-8
from django.urls import re_path

from rest_framework.routers import SimpleRouter

from concrete_datastore.concrete.meta import list_of_meta
from concrete_datastore.api.v1.signals import (
    send_email_pre_save,
    send_email_post_save,
)
from concrete_datastore.api.v1.views import (
    LoginApiView,
    RegisterApiView,
    ResetPasswordApiView,
    ChangePasswordView,
    AccountMeApiView,
    RetrieveSecureTokenApiView,
    SecureLoginApiView,
    GenerateSecureTokenApiView,
    RetrieveSecureConnectCode,
    SecureLoginCodeApiView,
)
from concrete_datastore.api.v1 import views, DEFAULT_API_NAMESPACE

app_name = 'concrete_datastore.concrete'

if not callable(send_email_pre_save) or not callable(send_email_post_save):
    raise ValueError("The method used to send emails should be a callable")

# API Front end
router = SimpleRouter()

for meta_model in list_of_meta:
    if meta_model.get_model_name() in ["EntityDividerModel", "UndividedModel"]:
        continue

    viewset = getattr(
        views, '{}ModelViewSet'.format(meta_model.get_model_name())
    )

    router.register(
        prefix=meta_model.get_dashed_case_class_name(),
        viewset=viewset,
        basename=meta_model.get_dashed_case_class_name(),
    )


specific_urlpatterns = [
    re_path(
        r'^auth/login/',
        LoginApiView.as_view(api_namespace=DEFAULT_API_NAMESPACE),
        name='login-view',
    ),
    re_path(
        r'^auth/register/',
        RegisterApiView.as_view(api_namespace=DEFAULT_API_NAMESPACE),
        name='register-view',
    ),
    re_path(
        r'^account/me/',
        AccountMeApiView.as_view(api_namespace=DEFAULT_API_NAMESPACE),
        name='account-me',
    ),
    re_path(
        r'^auth/change-password/',
        ChangePasswordView.as_view(api_namespace=DEFAULT_API_NAMESPACE),
        name='change-password',
    ),
    re_path(
        r'^auth/reset-password/',
        ResetPasswordApiView.as_view(),
        name='reset-password',
    ),
    re_path(
        r'secure-connect/retrieve-token',
        RetrieveSecureTokenApiView.as_view(),
        name='retrieve-secure-token',
    ),
    re_path(
        r'secure-connect/retrieve-code/',
        RetrieveSecureConnectCode.as_view(),
        name='retrieve-secure-code',
    ),
    re_path(
        r'secure-connect/login/',
        SecureLoginApiView.as_view(api_namespace=DEFAULT_API_NAMESPACE),
        name='secure-connect-login',
    ),
    re_path(
        r'secure-connect/login-code/',
        SecureLoginCodeApiView.as_view(api_namespace=DEFAULT_API_NAMESPACE),
        name='secure-connect-login-code',
    ),
    re_path(
        r'secure-connect/generate-token',
        GenerateSecureTokenApiView.as_view(),
        name='generate-secure-token',
    ),
]

urlpatterns = router.urls + specific_urlpatterns



# coding: utf-8
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST


class ConcreteBadResponse(Response):
    def __init__(self, message):
        super().__init__(
            data={"message": message, "_errors": ["INVALID_DATA"]},
            status=HTTP_400_BAD_REQUEST,
        )



# coding: utf-8
import pendulum

api_format = 'YYYY-MM-DDTHH:mm:ss.SSSSSS\Z'


def format_datetime(dt):
    return ensure_pendulum(dt).format(api_format)


def ensure_pendulum(value):
    d = value
    if isinstance(d, str):
        d = pendulum.parse(d)
    if not isinstance(d, pendulum.Date):
        d = pendulum.instance(d)
    return d



# coding: utf-8
from importlib import import_module
from django.conf import settings


def is_mfa_enabled(user):
    module_name, func_name = settings.MFA_RULE_PER_USER.rsplit('.', 1)
    module = import_module(module_name)
    use_mfa_rule = getattr(module, func_name)
    #: If the user has a totp_device, the mfa is enabled
    return use_mfa_rule(user=user) or user.totp_device is not None



# coding: utf-8
from __future__ import unicode_literals, absolute_import, print_function

from concrete_datastore.parsers.constants import STD_SPECIFIER


def validate_specifier(specifier_dict):
    specifier_type = specifier_dict['std.specifier']
    specifier_keys = STD_SPECIFIER[specifier_type]

    for key in specifier_keys:
        if key not in specifier_dict:
            raise ValueError(f'Missing {key}')



# coding: utf-8
import yaml
import json
import base64
import qrcode
import qrcode.image.svg

from django.conf import settings
from django.http import (
    HttpResponse,
    JsonResponse,
    HttpResponseForbidden,
    StreamingHttpResponse,
)
from django.views.generic import TemplateView
from django.contrib.auth import get_user_model

from rest_framework.views import APIView
from rest_framework import authentication
from rest_framework.renderers import OpenAPIRenderer
from concrete_datastore.api.v1.authentication import (
    TokenExpiryAuthentication,
    URLTokenExpiryAuthentication,
)

import concrete_datastore
from concrete_datastore.interfaces.yaml_renderer import DatamodelYamlToHtml
from concrete_datastore.concrete.constants import MFA_OTP
from concrete_datastore.routes.forms import ConfigureOTPLoginForm
from concrete_datastore.interfaces.openapi_schema_generator import (
    SchemaGenerator,
)


def service_status_view(request):
    plugins = {}
    # INSTALLED_PLUGIN is a dict with the plugin/core as key and the
    # version as value
    if isinstance(settings.INSTALLED_PLUGINS, dict):
        plugins = settings.INSTALLED_PLUGINS
    data = {
        'version': concrete_datastore.__version__,
        'datamodel_version': settings.DATAMODEL_VERSION,
        'api': concrete_datastore.api.v1_1.__version__,
        'plugins': plugins,
        'healthy': True,
        'message': '',
        'name': 'concrete-datastore',
        'license': 'All rights reserved. Netsach 2021.',
    }
    if settings.USE_CORE_AUTOMATION:
        try:
            # ImportError if the import fails
            from ns_core.coreApp.models import (  # pylint: disable = import-error
                Parameter,
            )
        except ImportError:
            pass
        else:
            parameter, created = Parameter.objects.get_or_create(
                name='MAINTENANCE_MODE', defaults={'data': {'value': False}}
            )
            maintenance_mode = parameter.data.get('value', False)
            data['maintenance_mode'] = maintenance_mode

    return JsonResponse(data)


class DatamodelServer(APIView, TemplateView):
    template_name = "mainApp/datamodel.html"
    authentication_classes = (
        authentication.BasicAuthentication,
        authentication.SessionAuthentication,
        TokenExpiryAuthentication,
        URLTokenExpiryAuthentication,
    )

    def _get_datamodel_format(self, data_format='yaml'):
        datamodel_content_json = settings.META_MODEL_DEFINITIONS
        if data_format == 'json':
            return json.dumps(datamodel_content_json, indent=4)
        return yaml.dump(datamodel_content_json, allow_unicode=True)

    def get(self, request, *args, **kwargs):
        user = request.user
        if user.is_anonymous or user.is_at_least_admin is False:
            return HttpResponseForbidden()
        action = kwargs.get('action')
        data_format = request.GET.get('data-format', 'yaml').lower()
        if action is None:
            if data_format == 'json':
                return JsonResponse(settings.META_MODEL_DEFINITIONS)
            return HttpResponse(self._get_datamodel_format())
        if action == 'download':
            filename = f'datamodel.{data_format}'
            response = StreamingHttpResponse(
                self._get_datamodel_format(data_format=data_format),
                content_type=data_format,
            )
            response[
                'Content-Disposition'
            ] = 'attachment; filename="{}"'.format(filename)
            return response
        if action == 'view':
            return super().get(request, *args, **kwargs)
        return JsonResponse(
            data={'error': f'unknown action {action}'}, status=400
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        datamodel_content_json = settings.META_MODEL_DEFINITIONS
        context['json_content'] = json.dumps(datamodel_content_json, indent=2)
        datamodel_content_yml = self._get_datamodel_format()
        content = DatamodelYamlToHtml(datamodel_content_json, indent=2)
        context['yaml_displayed_content'] = content.render_yaml()
        context['yaml_content'] = datamodel_content_yml
        return context


class OpenApiView(APIView):
    patterns = None

    def get(self, request, spec_format):
        url_scheme = request._request.META['wsgi.url_scheme']
        try:
            http_host = request._request.META['HTTP_HOST']
        except KeyError:
            http_host = (
                f"{request._request.META['REMOTE_ADDR']}"
                f":{request._request.META['SERVER_PORT']}"
            )

        user_auth_header = request._request.headers.get('Authorization')
        if user_auth_header:
            user_token = user_auth_header.replace('Token ', '')
            from_db = True
        else:
            user_token = None
            from_db = False
        server = f'{url_scheme}://{http_host}'

        generator = SchemaGenerator(
            servers=[server],
            patterns=self.patterns,
            user_token=user_token,
            from_db=from_db,
        )
        schema = generator.get_schema(request=request)

        if spec_format == 'yaml':
            spec = OpenAPIRenderer().render(schema, renderer_context={})
            return HttpResponse(spec.decode())

        return JsonResponse(schema)


class ConfigureOTPView(TemplateView):
    template_name = 'mainApp/configure-otp.html'
    form_class = ConfigureOTPLoginForm
    success_url = '.'

    def post(self, request, *args, **kwargs):
        User = get_user_model()
        context = self.get_context_data()
        form = context['form']
        if form.is_valid():
            email = form.cleaned_data['email']
            user = User.objects.get(email=email)
            context['user_is_authenticated'] = True
            device, _ = user.emaildevice_set.get_or_create(
                mfa_mode=MFA_OTP, confirmed=True
            )
            img = qrcode.make(
                device.config_url, image_factory=qrcode.image.svg.SvgImage
            )

            base64_qrcode_img_bytes = base64.b64encode(img.to_string())
            base64_qrcode_img = base64_qrcode_img_bytes.decode('utf-8')
            context['base64_qrcode_img'] = base64_qrcode_img

        return super().render_to_response(context)

    def get_context_data(self, **kwargs):
        context = super(ConfigureOTPView, self).get_context_data(**kwargs)
        form = ConfigureOTPLoginForm(self.request.POST or None)
        context["form"] = form
        context["platform_name"] = settings.PLATFORM_NAME
        return context



# coding: utf-8

from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = (
        'Get db connection string, could be used with pg_dump,'
        '`pg_dump --dbname=$CONNECTION_STRING`'
        'see https://www.postgresql.org/docs/14/libpq-connect.html'
    )

    def get_pg_dump_params(self):
        db_host = settings.DATABASES['default']['HOST']
        db_port = settings.DATABASES['default']['PORT']
        db_user = settings.DATABASES['default']['USER']
        db_name = settings.DATABASES['default']['NAME']
        db_pwd = settings.DATABASES['default']['PASSWORD']
        return f'postgresql://{db_user}:{db_pwd}@{db_host}:{db_port}/{db_name}'

    def handle(self, *args, **options):
        self.stdout.write(self.get_pg_dump_params())



from django.conf import settings
from rest_framework.throttling import UserRateThrottle, AnonRateThrottle
from django.contrib.auth.models import AbstractUser, AnonymousUser


class CustomUserRateThrottle(UserRateThrottle):
    def get_cache_key(self, request, view):
        if settings.ENABLE_AUTHENTICATED_USER_THROTTLING is False:
            return None
        user = request.user
        if not isinstance(user, (AbstractUser, AnonymousUser)):
            return None
        if user.is_authenticated:
            return user.email
        return None

    def get_rate(self):
        return settings.USER_THROTTLING_RATE


class CustomAnonymousRateThrottle(AnonRateThrottle):
    def get_cache_key(self, request, view):
        if settings.ENABLE_ANONYMOUS_USER_THROTTLING is False:
            return None
        user = request.user
        if not isinstance(user, (AbstractUser, AnonymousUser)):
            return None

        if user.is_authenticated:
            return None
        return super().get_cache_key(request, view)

    def get_rate(self):
        return settings.ANONYMOUS_THROTTLING_RATE



# coding: utf-8
ATLEAST_LEVEL_ATTRS = {
    'simpleuser': {'is_active': True},
    'manager': {'is_staff': True},
    'admin': {'admin': True},
    'superuser': {'is_superuser': True},
}

EXACT_LEVEL_ATTRS = {
    'blocked': {
        'is_active': False,
        'is_staff': False,
        'admin': False,
        'is_superuser': False,
    },
    'simpleuser': {
        'is_active': True,
        'is_staff': False,
        'admin': False,
        'is_superuser': False,
    },
    'manager': {
        'is_active': True,
        'is_staff': True,
        'admin': False,
        'is_superuser': False,
    },
    'admin': {
        'is_active': True,
        'is_staff': True,
        'admin': True,
        'is_superuser': False,
    },
    'superuser': {
        'is_active': True,
        'is_staff': True,
        'admin': True,
        'is_superuser': True,
    },
}

CRUD_LEVEL = ["anonymous", "authenticated", "admin", "superuser", "manager"]

LIST_USER_LEVEL = ["blocked", "simpleuser", "manager", "admin", "superuser"]

HANDELED_MODELISATION_VERSIONS = ('1.0.0',)

TYPE_EQ = {
    'ForeignKey': 'fk',
    'ManyToManyField': 'm2m',
    'BooleanField': 'bool',
    'TextField': 'txt',
    'CharField': 'char',
    'URLField': 'url',
    'IntegerField': 'int',
    'FloatField': 'float',
    'DecimalField': 'float',
    'JSONField': 'json',
    'FileField': 'file',
    'ImageField': 'file',
    'DateTimeField': 'datetime',
    'DateField': 'date',
    'BigIntegerField': 'int',
}

EMPTY_VALUES_MAP = {
    'TextField': {'empty_value': '', 'field_type': str},
    'CharField': {'empty_value': '', 'field_type': str},
}

MFA_OTP = "MFA_OTP"
MFA_EMAIL = "MFA_EMAIL"



# coding: utf-8

from django.apps import apps
from django.conf import settings
from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from concrete_datastore.concrete.meta import list_of_meta
from concrete_datastore.admin.admin_models import MetaUserAdmin, MetaAdmin
from concrete_datastore.admin.admin_site import get_admin_site
from concrete_datastore.concrete.models import (
    ConcreteRole,
    AuthToken,
    ConcretePermission,
    EmailDevice,
)
from concrete_datastore.concrete.models import (
    divider,
    divider_field_name,
    UNDIVIDED_MODEL,
)
from concrete_datastore.interfaces.csv import csv_streaming_response


main_app = apps.get_app_config('concrete')

admin_site = get_admin_site()

admin_site.site_header = settings.ADMIN_HEADER
admin_site.site_url = None
for meta_model in list_of_meta:
    model_name = meta_model.get_model_name()
    if model_name in ["EntityDividerModel", "UndividedModel"]:
        continue

    models_fields = ('uid',)
    m2m_fields = []
    attrs = {}

    for field_name, field in meta_model.get_fields():
        models_fields += (field_name,)
        if field.f_type == 'ManyToManyField':
            m2m_fields.append(field_name)

    if model_name == 'User':
        list_display = (meta_model.get_property('m_list_display') or []) + [
            'level'
        ]
        #:  Model user necessarily have an email field. It should be
        #:  on top of the list display
        if 'email' in list_display:
            list_display.remove('email')
        list_display.insert(0, 'email')
        meta_user_admin_cls = MetaUserAdmin
        user_custom_fields = [
            field
            for field in models_fields
            if field not in MetaUserAdmin.user_fields
        ]
        user_custom_fields += ['external_auth']
        meta_user_admin_cls.fieldsets = (
            (_('User fields'), {'fields': user_custom_fields}),
        ) + meta_user_admin_cls.fieldsets
        ancestors = [meta_user_admin_cls]

    else:
        list_display = ['uid'] + (
            meta_model.get_property('m_list_display') or []
        )
        ancestors = [MetaAdmin]
        meta_fieldsets = MetaAdmin.fieldsets
        meta_model_fields = MetaAdmin.models_fields
        fieldsets = ((None, {'fields': models_fields}),) + meta_fieldsets
        if model_name != divider and model_name not in UNDIVIDED_MODEL:
            fieldsets += ((_('Scope'), {'fields': (divider_field_name,)}),)
        attrs.update({'fieldsets': fieldsets, 'filter_horizontal': m2m_fields})

    def make_list_filter(mm, mdl):
        def get_list_filter(self, request):
            initial_list_filter = mm.get_property('m_filter_fields') or []
            list_filter = ['creation_date', 'modification_date']
            for field_name in initial_list_filter:
                nb_objects = mdl.objects.values(field_name).distinct().count()
                if nb_objects <= settings.LIMIT_DEACTIVATE_FILTER_IN_ADMIN:
                    list_filter.append(field_name)
            return list_filter

        return get_list_filter

    def export_csv(self, request, queryset):
        fields = [f.name for f in queryset.model._meta.fields]
        return csv_streaming_response(
            request, queryset.values(*fields), fields
        )

    export_csv.short_description = 'Export CSV (UTF-8)'

    def make_actions():
        actions = ["export_csv"]
        return actions

    model = main_app.models[meta_model.get_model_name().lower()]

    attrs.update(
        {
            'list_display': list_display,
            'search_fields': meta_model.get_property('m_search_fields') or [],
            'get_list_filter': make_list_filter(meta_model, model),
            'actions': make_actions(),
            'export_csv': export_csv,
        }
    )
    admin_site.register(
        model,
        type(
            str('{}MetaAdmin'.format(meta_model.get_model_name())),
            tuple(ancestors),
            attrs,
        ),
    )


@admin.register(AuthToken, site=admin_site)
class AuthTokenAdmin(admin.ModelAdmin):
    ordering = ('expiration_date',)
    list_display = [
        'key',
        'user',
        'expired',
        'expiration_date',
        'last_action_date',
    ]
    search_fields = ['user__email', 'key']
    readonly_fields = ['key']
    list_filter = ['expiration_date', 'last_action_date', 'expired']
    fields = ['key', 'user', 'expired', 'expiration_date', 'last_action_date']


class SaveModelMixin:
    def save_model(self, request, obj, form, change):
        if change is False:
            obj.created_by = request.user
        super().save_model(request, obj, form, change)


@admin.register(ConcretePermission, site=admin_site)
class ConcretePermissionAdmin(SaveModelMixin, admin.ModelAdmin):
    list_display = ['uid', 'model_name', 'creation_date']
    search_fields = ['model_name']
    readonly_fields = [
        'uid',
        'created_by',
        'creation_date',
        'modification_date',
    ]
    date_hierarchy = 'creation_date'

    fields = [
        'uid',
        'model_name',
        'create_roles',
        'retrieve_roles',
        'update_roles',
        'delete_roles',
        'creation_date',
        'modification_date',
        'created_by',
    ]
    filter_horizontal = [
        'create_roles',
        'retrieve_roles',
        'update_roles',
        'delete_roles',
    ]


@admin.register(ConcreteRole, site=admin_site)
class ConcreteRoleAdmin(SaveModelMixin, admin.ModelAdmin):
    list_display = ['uid', 'name', 'creation_date']
    search_fields = ['name']
    readonly_fields = [
        'uid',
        'created_by',
        'creation_date',
        'modification_date',
    ]
    date_hierarchy = 'creation_date'

    fields = [
        'uid',
        'name',
        'users',
        'creation_date',
        'modification_date',
        'created_by',
    ]
    filter_horizontal = ['users']


@admin.register(EmailDevice, site=admin_site)
class EmailDeviceAdmin(SaveModelMixin, admin.ModelAdmin):
    list_display = [
        'uid',
        'user',
        'mfa_mode',
        'confirmed',
        'creation_date',
        'modification_date',
    ]
    search_fields = ['user__email']
    readonly_fields = [
        'uid',
        'user',
        'key',
        'creation_date',
        'modification_date',
        'created_by',
    ]
    date_hierarchy = 'creation_date'

    fields = [
        'uid',
        'user',
        'key',
        'confirmed',
        'mfa_mode',
        'digits',
        'step',
        't0',
        'drift',
        'creation_date',
        'modification_date',
    ]
    list_filter = ['mfa_mode', 'confirmed']



# coding: utf-8
from django.urls import re_path, include
from django.conf import settings
from django.views.static import serve
from django.views.generic import TemplateView
from django.contrib.auth import views as auth_views
from concrete_datastore.admin.admin import admin_site
from .views import (
    service_status_view,
    OpenApiView,
    DatamodelServer,
    ConfigureOTPView,
)

app_name = 'concrete_datastore.concrete'

api_v1_1_urls = re_path(
    r'^api/v1\.1/',
    include('concrete_datastore.api.v1_1.urls', namespace='api_v1_1'),
)
api_v1_urls = re_path(
    r'^api/v1/', include('concrete_datastore.api.v1.urls', namespace='api_v1')
)

#:  Swagger and OpenAPI should be available on if
#:    - DEBUG is True
#:    - DEBUG is False and ENABLE_SWAGGER_UI is True
#:
#:  | DEBUG | ENABLE_SWAGGER_UI | result |
#:  | ----- | ----------------- | ------ |
#:  |  0    |        0          |   0    |
#:  |  0    |        1          |   1    |
#:  |  1    |        0          |   1    |
#:  |  1    |        1          |   1    |
swagger_urls = []
if settings.DEBUG or settings.ENABLE_SWAGGER_UI:
    swagger_urls = [
        re_path(
            fr'{settings.SWAGGER_SPEC_PATH}\.(?P<spec_format>json|yaml)$',
            OpenApiView.as_view(patterns=[api_v1_1_urls]),
            name='openapi-schema',
        ),
        re_path(
            fr'^{settings.SWAGGER_UI_PATH}/',
            TemplateView.as_view(
                template_name='mainApp/swagger-ui.html',
                extra_context={'schema_url': 'openapi-schema'},
            ),
            name='swagger-ui',
        ),
    ]


urlpatterns = [
    re_path(
        r'^auth/configure-otp/',
        ConfigureOTPView.as_view(),
        name='configure-otp',
    ),
    re_path(r'^oauth/', include('social_django.urls', namespace='social')),
    re_path(r'^status/$', service_status_view, name='service-status-view'),
    re_path(
        r'^datamodel/(?P<action>download|view/?)?$',
        DatamodelServer.as_view(),
        name='datamodel',
    ),
    re_path(
        r'^c/',
        include('concrete_datastore.concrete.urls', namespace='concrete'),
    ),
    api_v1_urls,
    api_v1_1_urls,
    *swagger_urls,
]
if settings.ADMIN_URL_ENABLED:
    urlpatterns += [
        re_path(rf'^{settings.ADMIN_ROOT_URI}/', admin_site.urls, name='admin')
    ]

if settings.DEBUG:
    try:
        import debug_toolbar

        urlpatterns += [re_path(r'^__debug__/', include(debug_toolbar.urls))]
    except ImportError:
        # Ignore debug_toolbar if not installed
        pass

    urlpatterns += [
        re_path(
            r'^m/(?P<path>.*)$', serve, {'document_root': settings.MEDIA_ROOT}
        ),
        re_path(
            r'^s/(?P<path>.*)$', serve, {'document_root': settings.STATIC_ROOT}
        ),
    ]
if settings.USE_CORE_AUTOMATION:
    # ImportError if the import fails
    import ns_core  # pylint: disable = import-error
    from ns_core.coreApp.admin_site import (  # pylint: disable = import-error
        admin_site as core_admin_site,
    )

    urlpatterns += [
        re_path(r'^core/', include('ns_core.coreApp.urls', namespace='')),
        re_path(r'^core-admin/', core_admin_site.urls, name='core-admin'),
    ]


urlpatterns += [
    re_path(
        r'^$',
        TemplateView.as_view(template_name='mainApp/index.html'),
        name='index',
    ),
    re_path(
        r'^oauth-logged$',
        TemplateView.as_view(template_name='mainApp/logged-using-oauth.html'),
        name='index-logged',
    ),
]



# coding: utf-8
from concrete_datastore.parsers.constants import (
    VERSIONS_ATTRIBUTES,
    AUTHORIZED_IP_PROTOCOLS,
)


class ModelManagerGenericException(Exception):
    """
    Main Exception class to handle errors in datamodel
    These exceptions are related to the version 1.0.0 of the datamodel
    definition. For further versions, adequate exceptions must be added
    """

    version = '1.0.0'
    code = ''
    message = ''

    def __str__(self):
        return f'[{self.code}]: {self.message}'


class MissingKeyForDefinition(ModelManagerGenericException):
    code = 'MISSING_KEY'

    def __init__(self, key, keys_list, resource_type, *args, **kwargs):
        self.message = (
            f'Cannot find key "{key}" for some {resource_type}. '
            f'Please make sure all your {resource_type.lower()}s definitions '
            f'contain at least the following keys: {keys_list}.'
        )


class UnknownDatamodelVersionError(ModelManagerGenericException):
    code = 'WRONG_VERSION'
    message = 'Key "version" of manifest should be 1.0.0.'


class UnknownIPProtocol(ModelManagerGenericException):
    code = 'UnknownIPProtocol'

    def __init__(self, protocol, source, field_name, *args, **kwargs):
        self.message = (
            f"Unknow protocol '{protocol}' for model {source}"
            f" and field {field_name}. "
            f"Authorized protocols are {AUTHORIZED_IP_PROTOCOLS}"
        )


class MissingRelationForModel(ModelManagerGenericException):
    code = 'MISSING_RELATION'

    def __init__(self, rel_type, source, target, field_name, *args, **kwargs):
        self.message = (
            f'Missing {rel_type} from {source} to {target} for field '
            f'"{field_name}". Please add it to {rel_type} section of your '
            'datamodel.'
        )


class DuplicatedRelationForModel(ModelManagerGenericException):
    code = 'DUPLICATED_RELATION'

    def __init__(self, rel_type, source, field_name, target, *args, **kwargs):
        self.message = (
            f'Relation duplicated in {rel_type} from "{source}" to "{target}" '
            f'for field "{field_name}" Please remove the duplicated relations '
            'to make sure only one exists.'
        )


class MissingPermissionsOrQueriesForModel(ModelManagerGenericException):
    def __init__(self, param_type, model_name, *args, **kwargs):
        self.code = f'MISSING_{param_type.upper()}'
        self.message = (
            f'Model "{model_name}" does not have any {param_type}. '
            'Please specify one for this model in your datamodel.'
        )


class DuplicatedPermissionsOrQueriesForModel(ModelManagerGenericException):
    def __init__(self, param_type, model_name, *args, **kwargs):
        self.code = f'DUPLICATED_{param_type.upper()}'
        self.message = (
            f'Duplicated {param_type} for model "{model_name}". '
            'Please make sure it is unique in your datamodel.'
        )


class NameNotAllowed(ModelManagerGenericException):
    code = 'NAME_NOT_ALLOWED'

    def __init__(self, name, resource_type, *args, **kwargs):
        self.message = (
            f'"{name}" cannot be used as a {resource_type} name. Please '
            f'select another name for this {resource_type} in your datamodel.'
        )


class DuplicatedFieldsError(ModelManagerGenericException):
    code = 'DUPLICATED_FIELDS'

    def __init__(self, field_name, model_name, *args, **kwargs):
        self.message = (
            f'Duplicated field "{field_name}" in model "{model_name}". '
            'Fields must have unique names. Please rename the duplicates.'
        )


class DuplicatedReverseForModel(ModelManagerGenericException):
    code = 'DUPLICATED_REVERSE'

    def __init__(self, name, reverse_name, *args, **kwargs):
        self.message = (
            f'Model "{name}" has more than one relation with the same reverse '
            f'name ({reverse_name}). Reverses should have a unique name for '
            'the same model. Please rename the duplicates.'
        )


class UnknownDatatypeForField(ModelManagerGenericException):
    code = 'UNKNOWN_DATATYPE'

    def __init__(self, field_name, model_name, datatype, *args, **kwargs):
        allowed_datatypes = list(
            VERSIONS_ATTRIBUTES[self.version]['equivalence_table'].keys()
        )
        self.message = (
            f'Field "{field_name}" of model "{model_name}" has an invalid '
            f'datatype ({datatype}). Please choose a valid datatype '
            f'from {allowed_datatypes}.'
        )


class ProtectedFieldNameError(ModelManagerGenericException):
    code = 'FIELD_NAME_NOT_ALLOWED'

    def __init__(self, field_name, model_name, *args, **kwargs):
        self.message = (
            f'"{field_name}" is a protected field for model "{model_name}" '
            'and cannot be redefined. Please rename it.'
        )


class ProtectedModelNameError(ModelManagerGenericException):
    code = 'MODEL_NAME_NOT_ALLOWED'

    def __init__(self, model_name, *args, **kwargs):
        self.message = (
            f'"{model_name}" is a protected model. You cannot use '
            'this name to define a new model. Please rename it.'
        )


class DuplicatedModelError(Exception):
    """
    This exception is deprecated and will be removed in further versions
    """

    pass



# coding: utf-8
from django.conf import settings

from rest_framework import pagination, response


class ExtendedPagination(pagination.PageNumberPagination):
    def get_paginated_response(self, data):
        return response.Response(
            {
                'objects_count': self.page.paginator.count,
                'next': self.get_next_link(),
                'previous': self.get_previous_link(),
                'results': data,
                'objects_count_per_page': self.page.paginator.per_page,
                'num_total_pages': self.page.paginator.num_pages,
                'num_current_page': self.page.number,
                'max_allowed_objects_per_page': settings.API_MAX_PAGINATION_SIZE,
            }
        )

    def get_page_size(self, request):
        default_page_size = settings.REST_FRAMEWORK.get(
            'PAGE_SIZE', settings.API_MAX_PAGINATION_SIZE
        )
        try:
            page_size = int(
                request.GET.get('c_resp_page_size', default_page_size)
            )
        except ValueError:
            page_size = default_page_size

        if request.GET.get('c_resp_nested', 'true') == 'true':
            page_size = min(settings.API_MAX_PAGINATION_SIZE_NESTED, page_size)
        else:
            page_size = min(settings.API_MAX_PAGINATION_SIZE, page_size)
        return max(1, page_size)



# coding: utf-8
from django import forms as django_forms
from django.contrib.auth import forms
from django.utils.translation import ngettext_lazy, gettext_lazy as _
from django.contrib.auth.models import User
from django_otp.forms import OTPAuthenticationFormMixin
from django.contrib.auth import get_user_model
from concrete_datastore.concrete.constants import MFA_EMAIL


class MyAuthForm(forms.AuthenticationForm):
    username = django_forms.EmailField(
        label=_("Email Address"), max_length=254
    )


class OTPAuthenticationForm(MyAuthForm, OTPAuthenticationFormMixin):
    otp_error_messages = {
        'token_required': _('Please enter your OTP token.'),
        'challenge_exception': _('Error generating challenge: {0}'),
        'challenge_message_email': _(
            'Enter the two authentication code you received by email'
        ),
        'challenge_message_otp': _(
            'Enter the two-factor authentication code from a configured OTP application'
        ),
        'invalid_token': _(
            'Invalid token. Please make sure you have entered it correctly.'
        ),
        'n_failed_attempts': ngettext_lazy(
            "Verification temporarily disabled because of %(failure_count)d failed attempt, please try again soon.",
            "Verification temporarily disabled because of %(failure_count)d failed attempts, please try again soon.",
            "failure_count",
        ),
        'verification_not_allowed': _(
            "Verification of the token is currently disabled"
        ),
    }

    otp_token = django_forms.CharField(
        required=False,
        widget=django_forms.TextInput(attrs={'autocomplete': 'off'}),
    )

    # This is a placeholder field that allows us to detect when the user clicks
    # the otp_challenge submit button.
    otp_challenge = django_forms.CharField(required=False)

    def clean(self):
        self.cleaned_data = super().clean()
        self.clean_otp(self.get_user())

        return self.cleaned_data

    def _handle_challenge(self, device):
        challenge_message_name = 'challenge_message_otp'
        if device.mfa_mode == MFA_EMAIL:
            challenge_message_name = 'challenge_message_email'
            try:
                device.generate_challenge()
            except Exception as e:
                raise django_forms.ValidationError(
                    self.otp_error_messages['challenge_exception'].format(e),
                    code='challenge_exception',
                )

        #:  If the challenge is successfully generated and in order to show
        #:  the appropriate message, a `ValidationError` must be raised.
        #:  *NB* This validation error does not mean that an error occured
        raise django_forms.ValidationError(
            self.otp_error_messages[challenge_message_name],
            code=challenge_message_name,
        )

    def clean_otp(self, user):
        if user is None:
            return

        device = self._chosen_device(user)
        token = self.cleaned_data.get('otp_token')
        username = self.cleaned_data.get('username')
        password = self.cleaned_data.get('password')

        user.otp_device = None
        try:
            if token:
                user.otp_device = self._verify_token(user, token, device)
                user._is_verified = True
                user.save()
            elif username and password:
                self._handle_challenge(device)
            else:
                #: If no username or no password are given, the user
                #: would be None, so this case should never occur,
                #: but we keep it in case it happens
                raise django_forms.ValidationError(  # pragma: no cover
                    self.otp_error_messages['token_required'],
                    code='token_required',
                )
        finally:
            if user.otp_device is None:
                self._update_form(user)

    def _chosen_device(self, user):
        device = user.totp_device
        if device is None:
            device = user.emaildevice_set.filter(confirmed=True).first()
        if device is None:
            device = user.emaildevice_set.create(
                email=user.email, name='User default email', confirmed=True
            )
        user.emaildevice_set.exclude(pk=device.pk).delete()

        return device

    @staticmethod
    def device_choices(user):
        user_divices = user.emaildevice_set.filter(confirmed=True)
        if not user_divices:
            user_divices = [
                user.emaildevice_set.create(
                    email=user.email, name='User default email', confirmed=True
                )
            ]
        return list((d.persistent_id, d.name) for d in user_divices)


class MyCreationUserForm(forms.UserCreationForm):
    email = django_forms.EmailField(required=True)

    class Meta:
        model = User
        fields = ("email", "password1", "password2")

    def clean_email(self):
        email = self.cleaned_data.get("email").lower()
        Concreteuser = get_user_model()
        if Concreteuser.objects.filter(email=email).count() > 0:
            raise django_forms.ValidationError(
                _("A user with this email already exists.")
            )
        return email

    def save(self, commit=True):
        user = super(MyCreationUserForm, self).save(commit=False)
        user.email = self.cleaned_data['email'].lower()
        if commit:
            user.save()
        return user


class MyChangeUserForm(forms.UserChangeForm):
    MY_CHOICES = (
        ('SuperUser', 'Super User'),
        ('Admin', 'Admin'),
        ('Manager', 'Manager'),
        ('SimpleUser', 'Simple User'),
        ('Blocked', 'Blocked'),
    )

    user_level = django_forms.ChoiceField(choices=MY_CHOICES)

    def __init__(self, *args, **kwargs):
        obj = kwargs.get('instance')
        if obj is not None:
            user_level = ""
            if obj.is_superuser:
                user_level = 'SuperUser'
            elif obj.admin:
                user_level = 'Admin'
            elif obj.is_staff:
                user_level = 'Manager'
            elif obj.is_active:
                user_level = 'SimpleUser'
            else:
                user_level = 'Blocked'
            kwargs['initial'] = {'user_level': user_level}

        super(MyChangeUserForm, self).__init__(*args, **kwargs)



# coding: utf-8
from __future__ import unicode_literals, absolute_import, print_function


class FieldDescriptor:
    """
    Field descriptor base class
    """

    def __init__(self, name, *args, **kwargs):
        self.name = name

    def __get__(self, instance, cls=None):
        return instance.__dict__.get(self.name)

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value


class RelFieldDescriptor(FieldDescriptor):
    """
    Field descriptor for related data (rel_single)
    """

    pass


class RelIterableFieldDescriptor(FieldDescriptor):
    """
    Field descriptor for related data (rel_single)
    """

    def __get__(self, instance, cls=None):
        if instance.__dict__.get(self.name) is None:
            instance.__dict__[self.name] = set([])

        return instance.__dict__[self.name]

    def __set__(self, instance, value):
        raise AttributeError(
            'Cannot use set method. Use .add(*objects) and .remove(*objects)'
        )


field_descriptors = {
    'data': FieldDescriptor,
    'rel_single': RelFieldDescriptor,
    'rel_iterable': RelIterableFieldDescriptor,
}


def get_field_descriptor(f_type):
    if f_type == 'fk':
        return RelFieldDescriptor
    if f_type == 'm2m':
        return RelIterableFieldDescriptor
    return FieldDescriptor
